# pylint: disable=C0302,R0912
"""
@file
@brief Implements a class able to compute the predictions
from on an :epkg:`ONNX` model.
"""
from collections import OrderedDict
from io import BytesIO
from time import perf_counter
import warnings
import textwrap
import pprint
from keyword import iskeyword
import numpy
from scipy.sparse import coo_matrix
from onnx import (
    load, load_model, shape_inference,
    ModelProto, GraphProto, FunctionProto)
from onnx.helper import make_model
from ..tools.code_helper import make_callable, print_code
from ..onnx_tools.model_checker import check_onnx
from ..onnx_tools.onnx2py_helper import (
    _var_as_dict, numpy_min, numpy_max)
from ..onnx_tools.onnx_manipulations import (
    select_model_inputs_outputs, enumerate_model_node_outputs,
    overwrite_opset, insert_results_into_onnx)
from ..onnx_tools.optim import onnx_remove_node_unused
from .onnx_inference_node import OnnxInferenceNode
from .onnx_inference_exports import OnnxInferenceExport
from .onnx_shape_inference import OnnxShapeInference
from .ops_shape.shape_excs import (
    ShapeInferenceMissing, NotImplementedShapeInferenceError,
    ShapeInferenceException, ShapeInferenceDimensionError)


class OnnxInference:
    """
    Loads an :epkg:`ONNX` file or object or stream.
    Computes the output of the :epkg:`ONNX` graph.
    Several runtimes are available.

    * ``'python'``: the runtime implements every onnx operator
      needed to run a :epkg:`scikit-learn` model by using :epkg:`numpy`
      or C++ code.
    * ``'python_compiled'``: it is the same runtime than the previous
      one except every operator is called from a compiled function
      (@see me _build_compile_run) instead for a method going through
      the list of operator
    * ``'onnxruntime1'``: uses :epkg:`onnxruntime` (or `onnxruntime1-cuda`, ...)
    * ``'onnxruntime2'``: this mode is mostly used to debug as
      python handles calling every operator but :epkg:`onnxruntime`
      is called for every of them, this process may fail due to
      wrong inference type specially of the graph includes
      custom nodes, in that case, it is better to compute the output
      of intermediates nodes. It is much slower as fo every output, every
      node is computed but more robust.

    :param onnx_or_bytes_or_stream: :epkg:`onnx` object,
        bytes, or filename or stream
    :param runtime: runtime options
    :param skip_run: do not build the runtime
    :param inplace: use inplace computation as much as possible
    :param input_inplace: the computation is allowed
        to overwrite the input, see :meth:`_guess_inplace
        <mlprodict.onnxrt.onnx_inference.OnnxInference._guess_inplace>`
    :param ir_version: if not None, overwrite the default version
    :param target_opset: used to overwrite *target_opset*
    :param runtime_options: specific options for the runtime
    :param inside_loop: tells the runtime the graph is meant to
        be repeated multiple times (in that case, inputs and
        outputs may share the same name)
    :param static_inputs: Loop can use static variables,
        variables from the graph which runs the loop
        (enumerate of strings)
    :param new_outputs: if the loading fails, it might worth
        cutting the graph, if not None, the graph will
        be cut to have these new_outputs as the final outputs
    :param new_opset: overwrite the main opset and replaces
        by this new one
    :param existing_functions: a model may contain several local functions,
        this parameter is used when a local function is calling another
        local function previously defined.

    Among the possible runtime_options, there are:
    * *enable_profiling*: enables profiling for :epkg:`onnxruntime`
    * *session_options*: an instance of *SessionOptions* from
        :epkg:`onnxruntime`
    * *ir_version*: change ir_version

    .. versionchanged:: 0.9
        Parameters *existing_functions* was added.
        Removes *device* parameter. See runtime.
        Runtime `onnxruntime1-cuda` was added.
    """

    def __init__(self, onnx_or_bytes_or_stream, runtime=None,
                 skip_run=False, inplace=True,
                 input_inplace=False, ir_version=None,
                 target_opset=None, runtime_options=None,
                 session_options=None, inside_loop=False,
                 static_inputs=None, new_outputs=None, new_opset=None,
                 existing_functions=None):
        if isinstance(onnx_or_bytes_or_stream, bytes):
            self.obj = load_model(BytesIO(onnx_or_bytes_or_stream))
        elif isinstance(onnx_or_bytes_or_stream, BytesIO):
            self.obj = load_model(onnx_or_bytes_or_stream)
        elif isinstance(onnx_or_bytes_or_stream, str):
            self.obj = load(onnx_or_bytes_or_stream)
        elif hasattr(onnx_or_bytes_or_stream, 'graph'):
            self.obj = onnx_or_bytes_or_stream
        elif isinstance(onnx_or_bytes_or_stream, GraphProto):
            self.obj = make_model(onnx_or_bytes_or_stream,
                                  producer_name='mlprodict')
        elif isinstance(onnx_or_bytes_or_stream, FunctionProto):
            self.obj = onnx_or_bytes_or_stream
        else:
            raise TypeError("Unable to handle type {}.".format(  # pragma: no cover
                type(onnx_or_bytes_or_stream)))
        if ir_version is not None:
            self.obj.ir_version = ir_version
        if new_outputs is not None:
            self.obj = select_model_inputs_outputs(
                self.obj, outputs=new_outputs, infer_shapes=True)
        if new_opset is not None:
            self.obj = overwrite_opset(self.obj, new_opset)

        self.runtime = runtime
        self.skip_run = skip_run
        self.input_inplace = input_inplace
        self.inplace = inplace
        self.force_target_opset = target_opset
        self.runtime_options = runtime_options
        self.inside_loop = inside_loop
        self.static_inputs = static_inputs
        self._init(existing_functions)

    def __getstate__(self):
        """
        To pickle the object.
        """
        return {'onnx': self.obj.SerializeToString(),
                'runtime': self.runtime,
                'runtime_options': self.runtime_options,
                'skip_run': self.skip_run,
                'input_inplace': self.input_inplace,
                'inplace': self.inplace,
                'force_target_opset': self.force_target_opset,
                'static_inputs': self.static_inputs,
                'inside_loop': self.inside_loop}

    def __setstate__(self, state):
        """
        To unpickle the object.
        """
        onx = state['onnx']
        self.obj = load_model(BytesIO(onx))
        self.runtime = state['runtime']
        self.runtime_options = state['runtime_options']
        self.skip_run = state['skip_run']
        self.input_inplace = state['input_inplace']
        self.inplace = state['inplace']
        self.force_target_opset = state['force_target_opset']
        self.static_inputs = state['static_inputs']
        self.inside_loop = state['inside_loop']
        self._init()

    def _init(self, existing_functions=None):
        """
        Prepares the instance to deliver predictions.
        """
        self.graph_ = self.to_sequence(existing_functions)
        self.functions_ = self.graph_['functions']
        self.outputs_ = self.graph_['outputs']
        self.inputs_ = self.graph_['inputs']
        self.attributes_ = self.graph_['attributes']
        is_function_proto = isinstance(self.obj, FunctionProto)
        if is_function_proto:
            obj_graph = self.obj
        else:
            obj_graph = self.obj.graph

        for ino in [obj_graph.input, obj_graph.output]:
            for xy in ino:
                if isinstance(xy, str):
                    shape = None
                else:
                    shape = xy.type.tensor_type.shape
                    for d in shape.dim:
                        if (d.dim_value == 0 and "0" in str(d) and
                                'dim_param' not in str(d)):
                            if len(shape.dim) <= 1:
                                shape = None
                                break
                            # d.dim_value returns 0 whether is is 0 or empty.
                            # it may be a parameter as well
                            raise RuntimeError(  # pragma: no cover
                                "Wrong ONNX file, one input or output has "
                                "an empty shape: {}.".format(xy))

        self.target_opset_ = self.graph_['targets']
        if self.force_target_opset is not None:
            if isinstance(self.force_target_opset, dict):
                self.target_opset_ = self.force_target_opset  # pragma: no cover
            else:
                self.target_opset_ = {'': self.force_target_opset}
        self.ir_version_ = self.graph_['ir_version']

        if not self.skip_run:
            if self.runtime is not None and self.runtime.startswith('onnxruntime1'):
                # Loads the onnx with onnxruntime as a single file.
                del self.graph_
                from .ops_whole.session import OnnxWholeSession
                self._whole = OnnxWholeSession(
                    self.obj, self.runtime, self.runtime_options)
                self._run = self._run_whole_runtime
            else:
                self.sequence_ = self.graph_['sequence']
                self.inits_ = self.graph_['inits']
                self.statics_ = self.graph_['statics']
                dtype = self._guess_input_dtype()
                variables = self.inits_.copy()
                for node in self.sequence_:
                    domain = node.onnx_node.domain
                    target_opset = self.target_opset_.get(domain, None)
                    keyf = domain, node.onnx_node.op_type
                    if keyf in self.functions_:
                        node.setup_runtime(self.graph_['functions'][keyf])
                    elif self.runtime in ('onnxruntime2', 'empty'):
                        node.setup_runtime(
                            self.runtime, variables, self.__class__,
                            target_opset=target_opset, dtype=dtype,
                            domain=domain, ir_version=self.ir_version_,
                            runtime_options=self.runtime_options,
                            existing_functions=self.functions_,
                            build_inference_node_function=lambda fct:
                                OnnxInference(
                                    fct, runtime=self.runtime,
                                    skip_run=self.skip_run,
                                    inplace=self.inplace,
                                    runtime_options=self.runtime_options,
                                    inside_loop=self.inside_loop,
                                    static_inputs=self.static_inputs))
                    else:
                        node.setup_runtime(
                            self.runtime, variables, self.__class__,
                            target_opset=target_opset, domain=domain,
                            ir_version=self.ir_version_,
                            runtime_options=self.runtime_options,
                            existing_functions=self.functions_,
                            build_inference_node_function=lambda fct:
                                OnnxInference(
                                    fct, runtime=self.runtime,
                                    skip_run=self.skip_run,
                                    inplace=self.inplace,
                                    runtime_options=self.runtime_options,
                                    inside_loop=self.inside_loop,
                                    static_inputs=self.static_inputs))
                    if hasattr(node, 'ops_') and hasattr(node.ops_, 'typed_outputs_'):
                        for k, v in node.ops_.typed_outputs_:
                            variables[k] = v
                self._run = self._run_sequence_runtime

        if not self.skip_run and self.runtime in ('python', None):
            if is_function_proto:
                self.shapes_ = None
            else:
                self.shapes_ = self._set_shape_inference_runtime()
            if self.inplace:
                self.inplaces_ = self._guess_inplace(self.input_inplace)

        self.exporters_ = OnnxInferenceExport(self)
        self.to_json = self.exporters_.to_json
        self.to_dot = self.exporters_.to_dot
        self.to_python = self.exporters_.to_python
        self.to_text = self.exporters_.to_text
        self.to_onnx_code = self.exporters_.to_onnx_code

        if self.runtime in ('python_compiled', 'python_compiled_debug'):
            # switch the inference method to the compiled one
            _, fct, code = self._build_compile_run('debug' in self.runtime)
            setattr(self, '_run_compiled', fct)
            setattr(self, '_run_compiled_code', code)
            self._run = self._run_sequence_runtime_compiled

    def _run_sequence_runtime_compiled(
            self, inputs, clean_right_away=False, intermediate=False,
            verbose=0, node_time=False, yield_ops=None, fLOG=None,
            context=None, attributes=None):
        """
        Executes a compiled version of @see me _run_sequence_runtime,
        compiled with method @see me _build_compile_run.
        Every parameter with a default value is ignored.
        Switch to ``runtime='python'`` to enable those.
        """
        try:
            return self._run_compiled(  # pylint: disable=E1101
                inputs, yield_ops=yield_ops, context=context,
                attributes=attributes)
        except NameError as e:
            raise RuntimeError(  # pragma: no cover
                "Unable to compute prediction due to %r. Code:\n%s"
                "" % (e, print_code(
                    self._run_compiled_code))) from e  # pylint: disable=E1101

    def _guess_input_dtype(self):
        for _, v in self.graph_['inputs'].items():
            if 'type' not in v:
                continue  # pragma: no cover
            t = v['type']
            if 'elem' not in t:
                continue
            if t['elem'] == 'double':
                return numpy.float64
        return numpy.float32

    def __str__(self):
        """
        usual
        """
        rows = ['OnnxInference(...)']
        if hasattr(self, '_run_compiled_code'):
            rows.append(
                textwrap.indent(
                    self._run_compiled_code, '    '))  # pylint: disable=E1101
        else:
            rows.append(textwrap.indent(str(self.obj), '    '))
        return "\n".join(rows)

    def __repr__(self):
        """
        usual
        """
        return "OnnxInference(...)"  # pragma: no cover

    def check_onnx(self):
        """
        Checks the model follow :epkg:`ONNX` conventions.
        """
        check_onnx(self.obj)

    def shape_inference(self):
        """
        Infers the shape of the outputs
        with :epkg:`onnx` package.

        @return     A new :epkg:`ONNX` graph which defined outputs.
        """
        return shape_inference.infer_shapes(self.obj)

    @property
    def input_names(self):
        """
        Returns the names of all inputs.
        It does not include the optional inputs.

        .. versionchanged:: 0.6
            The list does not include optional inputs anymore.
        """
        if hasattr(self.obj, 'graph'):
            inits = set(_.name for _ in self.obj.graph.initializer)
            return [_.name for _ in self.obj.graph.input if _.name not in inits]
        return list(self.obj.input)

    @property
    def input_names_shapes(self):
        """
        Returns the names and shapes of all inputs.
        This method assumes all inputs are tensors.
        It does not include the optional inputs.

        .. versionchanged:: 0.6
            The list does not include optional inputs anymore.
        """
        names = set(self.input_names)
        return [(_.name, _var_as_dict(_)['type']['shape'])
                for _ in self.obj.graph.input if _.name in names]

    @staticmethod
    def _get_type_property(info, prop):
        if prop in info:
            return info[prop]
        if 'kind' in info and info['kind'] == 'sequence':
            if prop == 'shape':
                return ('?', )
        raise NotImplementedError(  # pragma: no cover
            f"Unable to retrieve property {prop!r} from {info!r}.")

    @property
    def input_names_shapes_types(self):
        """
        Returns the names, shapes, types of all inputs.
        This method assumes all inputs are tensors.
        It does not include the optional inputs.

        .. versionchanged:: 0.6
            The list does not include optional inputs anymore.
        """
        f = OnnxInference._get_type_property
        names = set(self.input_names)
        if isinstance(self.obj, FunctionProto):
            return [(_.name, f(_var_as_dict(_)['type'], 'shape'),
                     f"tensor({f(_var_as_dict(_)['type'], 'elem')})")
                    for _ in self.obj.input if _.name in names]
        return [(_.name, f(_var_as_dict(_)['type'], 'shape'),
                 f"tensor({f(_var_as_dict(_)['type'], 'elem')})")
                for _ in self.obj.graph.input if _.name in names]

    @property
    def output_names(self):
        """
        Returns the names of all outputs.
        """
        if isinstance(self.obj, FunctionProto):
            return [_ for _ in self.obj.output]
        return [_.name for _ in self.obj.graph.output]

    @property
    def output_names_shapes(self):
        """
        Returns the names and shapes of all outputs.
        This method assumes all inputs are tensors.
        """
        f = OnnxInference._get_type_property
        if isinstance(self.obj, FunctionProto):
            return [(_, None) for _ in self.obj.output]
        return [(_.name, f(_var_as_dict(_)['type'], 'shape'))
                for _ in self.obj.graph.output]

    @property
    def output_names_shapes_types(self):
        """
        Returns the names, shapes, types of all outputs.
        This method assumes all inputs are tensors.
        It does not include the optional outputs.

        .. versionadd:: 0.7
        """
        names = set(self.output_names)
        f = OnnxInference._get_type_property
        if isinstance(self.obj, FunctionProto):
            return [(_, None) for _ in self.obj.graph.output if _ in names]
        return [(_.name, f(_var_as_dict(_)['type'], 'shape'),
                 f"tensor({f(_var_as_dict(_)['type'], 'elem')})")
                for _ in self.obj.graph.output if _.name in names]

    def global_index(self, name):
        """
        Maps every name to one integer to avoid using dictionaries
        when running the predictions.

        @param      name        outputs name
        @return                 integer
        """
        if not hasattr(self, '_global_index'):
            self._global_index = {}
        if name in self._global_index:
            return self._global_index[name]
        self._global_index[name] = len(self._global_index)
        return self._global_index[name]

    def to_sequence(self, existing_functions=None):
        """
        Produces a graph to facilitate the execution.

        One example:

        .. exref::
            :title: Convert ONNX into graph

            An example on how to convert an :epkg:`ONNX`
            graph into a graph.

            .. runpython::
                :showcode:
                :warningout: DeprecationWarning

                import pprint
                import numpy
                from mlprodict.npy.xop import loadop
                from mlprodict.onnxrt import OnnxInference

                OnnxAiOnnxMlLinearRegressor = loadop(
                    ('ai.onnx.ml', 'LinearRegressor'))

                pars = dict(coefficients=numpy.array([1., 2.]),
                            intercepts=numpy.array([1.]),
                            post_transform='NONE')
                onx = OnnxAiOnnxMlLinearRegressor(
                    'X', output_names=['Y'], **pars)
                model_def = onx.to_onnx(
                    {'X': pars['coefficients'].astype(numpy.float32)},
                    outputs={'Y': numpy.float32},
                    target_opset=12)
                oinf = OnnxInference(model_def)
                pprint.pprint(oinf.to_sequence())

            See an example of representation in notebook
            :ref:`onnxvisualizationrst`.
        """
        inits = {}
        variables = {}
        outputs = {}
        nodes = {}
        statics = {}
        targets = {}
        functions = {}
        attributes = {}
        if existing_functions is not None:
            functions.update(existing_functions)
        is_function_proto = isinstance(self.obj, FunctionProto)
        if is_function_proto and self.obj.attribute:
            for att in self.obj.attribute:
                attributes[att] = None

        for o in self.obj.opset_import:
            targets[o.domain] = o.version

        if (hasattr(self.obj, 'functions') and len(self.obj.functions) > 0 and
                (self.runtime is None or not
                    self.runtime.startswith('onnxruntime1'))):
            for fct in self.obj.functions:
                try:
                    oinf = OnnxInference(
                        fct, runtime=self.runtime,
                        skip_run=self.skip_run,
                        inplace=self.inplace,
                        runtime_options=self.runtime_options,
                        inside_loop=self.inside_loop,
                        static_inputs=self.static_inputs,
                        existing_functions=functions)
                except RuntimeError as e:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to instantiate function %r, %r." % (
                            fct.domain, fct.name)) from e
                functions[fct.domain, fct.name] = oinf

        # static variables
        if self.static_inputs is not None:
            for n in self.static_inputs:
                statics[n] = {'name': n}
                self.global_index(n)

        obj_graph = (
            self.obj if isinstance(self.obj, FunctionProto)
            else self.obj.graph)

        # inputs
        for obj in obj_graph.input:
            if is_function_proto:
                variables[obj] = {'name': obj}
                self.global_index(obj)
            else:
                variables[obj.name] = _var_as_dict(obj)
                self.global_index(obj.name)

        # outputs
        for obj in obj_graph.output:
            if is_function_proto:
                outputs[obj] = {'name': obj}
                self.global_index(obj)
            else:
                if hasattr(obj, 'type') and str(obj.type) != '':
                    outputs[obj.name] = _var_as_dict(obj)
                else:
                    outputs[obj.name] = {'name': obj.name}
                self.global_index(obj.name)

        # initializer
        if not is_function_proto:
            for obj in obj_graph.initializer:
                init_obj = _var_as_dict(obj)
                if init_obj is None:
                    raise RuntimeError(  # pragma: no cover
                        f"Unable to convert an initializer\n{obj}")
                inits[obj.name] = init_obj
                self.global_index(obj.name)
                if 'value' not in inits[obj.name]:
                    raise RuntimeError(  # pragma: no cover
                        "One initializer has no value: '{}'\n{}\n{}".format(
                            obj.name, inits[obj.name], obj))

        # nodes
        for node in obj_graph.node:
            dobj = _var_as_dict(node)
            if dobj is None:
                raise RuntimeError(  # pragma: no cover
                    f"Unable to convert a node\n{node}")
            if 'atts' in dobj:
                atts = dobj['atts']
                for k, v in atts.items():
                    if not isinstance(v, dict) or (
                            'value' not in v and 'ref_attr_name' not in v):
                        raise RuntimeError(  # pragma: no cover
                            "A parameter has no (sparse) value '{}' "
                            "for node '{}'\nv={}\ndobj=[{}]".format(
                                k, node.name, v, node))
            if node.name in nodes:  # pragma: no cover
                i = 2
                while True:
                    new_name = "%s_n%i" % (node.name, i)
                    if new_name not in nodes:
                        break
                    i += 1
            else:
                new_name = node.name
            nodes[new_name] = OnnxInferenceNode(node, dobj, self.global_index)

        # names
        names = {}
        for k, v in statics.items():
            if (k, 0) in names:
                raise RuntimeError(  # pragma: no cover
                    f"Static variables '{k}' already exists (tag='{names[k, 0][0]}').")
            names[k, 0] = ('S', v)
        for k, v in inits.items():
            if (k, 0) in names:
                raise RuntimeError(  # pragma: no cover
                    f"Initializer '{k}' already exists (tag='{names[k, 0][0]}').")
            names[k, 0] = ('C', v)
        for k, v in variables.items():
            if (k, 0) in names:
                if k in inits:
                    # Kind of default value for an input
                    continue
                raise RuntimeError(  # pragma: no cover
                    f"Variable '{k}' already exists (tag='{names[k, 0][0]}').")
            names[k, 0] = ('I', v)
        for k, v in outputs.items():
            if (k, 0) in names and (self.runtime != 'empty' and len(nodes) > 0):
                if not self.inside_loop or names[k, 0][0] != 'I':
                    raise RuntimeError(  # pragma: no cover
                        f"Output '{k}' already exists (tag='{names[k, 0][0]}').")
                else:
                    # For input, output sharing the same name, we marked the name
                    # as an input.
                    continue
            names[k, 0] = ('O', v)
        for k, v in nodes.items():
            if (k, 1) in names:
                raise RuntimeError(  # pragma: no cover
                    "Node '{}' already exists (tag='{}'). "
                    "Use inside_loop=True to bypass this exception.".format(
                        k, names[k, 0][0]))
            names[k, 1] = ('N', v)

        # ordering
        order = {}
        modif = 1
        intermediate = {}
        while modif > 0:
            modif = 0
            for (k, _), v in names.items():
                if (k, 1) in order:
                    # The operator node is already processed.
                    continue
                if v[0] in {'I', 'C', 'S'}:
                    if (k, 0) not in order:
                        order[k, 0] = len(order)  # A data node.
                        modif += 1
                    continue
                if v[0] == 'O':
                    continue
                if all((inp, 0) in order for inp in v[1].inputs if inp != ''):
                    # If all inputs are available,
                    # We tell the operator node is processed.
                    order[k, 1] = len(order)
                    modif += 1
                    for o in v[1].outputs:
                        if (o, 0) in order:
                            raise RuntimeError(  # pragma: no cover
                                "Two nodes share the same output '{}' "
                                "or an operator and an output "
                                "share the same name. "
                                "(node: {}).".format(o, v[1]))
                        # We add a data node.
                        order[o, 0] = len(order)
                        intermediate[o] = None
                        modif += 1

        # compute
        rev = [(v, k[0], k[1]) for k, v in order.items()]
        rev.sort()
        sequence = []
        for _, name, node_kind in rev:
            if name not in nodes:
                continue
            if node_kind == 0:
                # It is an output which shares the same name
                # as a node.
                continue
            node = nodes[name]
            node.set_order(len(sequence))
            sequence.append(node)

        # defines where an intermediare output is not needed
        last_used = {}
        for node in sequence:
            for inp in node.inputs:
                last_used[inp] = node.order
        for k, ord in last_used.items():
            sequence[ord].add_variable_to_clean(k)

        results = dict(inits=inits, inputs=variables, outputs=outputs,
                       attributes=attributes,
                       nodes=nodes, sequence=sequence,
                       functions=functions,
                       intermediate=intermediate,
                       targets=targets,
                       ir_version=(
                           None if is_function_proto
                           else self.obj.ir_version),
                       statics=statics)
        if len(sequence) < len(nodes):
            # Not all node will be executed.
            raise RuntimeError(  # pragma: no cover
                "Unable to run all nodes.\n--Nodes--\n%s\n--Sequence--\n%s"
                "\n--Inputs--\n%s\n--Inits--\n%s\n--Statics\n%s"
                "" % (pprint.pformat(nodes), pprint.pformat(sequence),
                      pprint.pformat(list(variables)),
                      pprint.pformat(list(inits)),
                      pprint.pformat(list(statics))))
        return results

    #############
    # inference #
    #############

    def run(self, inputs, clean_right_away=False,
            intermediate=False, verbose=0, node_time=False,
            overwrite_types=None, yield_ops=None, fLOG=None,
            context=None, attributes=None):
        """
        Computes the predictions for this :epkg:`onnx` graph.

        :param inputs: inputs as dictionary or a dataframe
        :param clean_right_away: clean the intermediate outputs
                as soon as they are not needed
        :param intermediate: returns a dictionary of intermediate
            variables instead of the results only
        :param verbose: display information while predicting
        :param node_time: measure time of each node
        :param overwrite_types: shape inference does not work all the time,
            this allows to force types when building intermediate
            results, see @see fn select_model_inputs_outputs
        :param yield_ops: dictionary to overwrite the output of
            operator *YieldOp*
        :param fLOG: logging function if *verbose > 0*
        :param context: local variables, needed when this object is a subgraph
        :param attributes: this uses when this class runs a :epkg:`FunctionProto`
            to store the values of the attributes of the function
        :return: outputs as dictionary
            and a second dictionary of the time spent
            in each node if *node_time* is True

        .. exref::
            :title: Computes predictions with any runtime

            The following example compares predictions
            between :epkg:`scikit-learn` and this runtime
            for the python runtime.

            .. runpython::
                :showcode:
                :warningout: DeprecationWarning

                import numpy
                from sklearn.linear_model import LinearRegression
                from sklearn.datasets import load_iris
                from sklearn.model_selection import train_test_split
                from mlprodict.onnxrt import OnnxInference
                from mlprodict.onnx_conv import to_onnx

                iris = load_iris()
                X, y = iris.data, iris.target
                X_train, X_test, y_train, _ = train_test_split(X, y)
                clr = LinearRegression()
                clr.fit(X_train, y_train)

                exp = clr.predict(X_test[:5])
                print(exp)

                model_def = to_onnx(clr, X_train.astype(numpy.float32),
                                    target_opset=12)
                oinf = OnnxInference(model_def)
                y = oinf.run({'X': X_test[:5]})
                print(y)

        The function returns all intermediate outputs
        if *intermediate* is True. In case of runtime
        *onnxruntime1*, if intermediate is True,
        the first class builds all :epkg:`ONNX` cut out
        to keep the one output and converted into
        *OnnxInference*.

        .. versionchanged:: 0.9
            Parameter *attributes* was added.
        """
        def retype(col_array):
            if (hasattr(col_array, 'categories') and
                    hasattr(col_array, 'from_codes')):
                # isinstance(col_array, pandas.Categorical):
                return col_array.astype(numpy.int64)
            return col_array

        if hasattr(inputs, 'columns') and hasattr(inputs, 'iloc'):
            # == isinstance(inputs, pandas.DataFrame)
            inputs = OrderedDict((
                name, retype(numpy.expand_dims(inputs[name].values, axis=1)))
                for name in inputs.columns)
        if intermediate:
            if self.inplace:
                raise RuntimeError(  # pragma: no cover
                    "inplace must be False if intermediate is True, a container "
                    "might be used by several nodes.")
            return self._run(inputs, clean_right_away=False,  # pylint: disable=E1123
                             intermediate=intermediate,
                             verbose=verbose, node_time=node_time,
                             overwrite_types=overwrite_types,
                             yield_ops=yield_ops, fLOG=fLOG,
                             context=context, attributes=attributes)
        if overwrite_types is not None:
            raise RuntimeError(  # pragma: no cover
                "overwrite_types is not used if intermediate is False.")
        return self._run(inputs, clean_right_away=False,  # pylint: disable=E1123
                         intermediate=intermediate,
                         verbose=verbose, node_time=node_time,
                         yield_ops=yield_ops, fLOG=fLOG,
                         context=context, attributes=attributes)

    def run2onnx(self, inputs, verbose=0, fLOG=None,
                 as_parameter=True, suffix='_DBG',
                 param_name=None, node_type='DEBUG',
                 domain='DEBUG', domain_opset=1,
                 attributes=None):
        """
        Executes the graphs with the given inputs, then adds the intermediate
        results into ONNX nodes in the original graph. Once saved, it can be
        looked with a tool such as :epkg:`netron`.

        :param inputs: inputs as dictionary or a dataframe
        :param verbose: display information while predicting
        :param fLOG: logging function if *verbose > 0*
        :param as_parameter: add new nodes with results as one parameter
            (True) or as initializer (False)
        :param suffix: suffix to add to new results
        :param param_name: name of the parameter to add
            (by default the result name), it can be a function
            `param_name(reult_name) -> parameter_name`
        :param node_type: type of the new node
        :param domain: domain the new node
        :param domain_opset: opset for *domain*
        :param attributes: values for attributes if this class runs a
            :epkg:`FunctionProto`
        :return: outputs as dictionary
            and the onnx graph with new nodes

        The following example shows how to use it.

        .. gdot::
            :script: DOT-SECTION

            from sklearn.linear_model import LinearRegression
            from sklearn.datasets import load_iris
            from mlprodict.onnxrt import OnnxInference
            import numpy

            iris = load_iris()
            X = iris.data[:, :2]
            y = iris.target
            lr = LinearRegression()
            lr.fit(X, y)

            from mlprodict.onnx_conv import to_onnx
            model_onnx = to_onnx(lr, X.astype(numpy.float32))
            oinf = OnnxInference(model_onnx, inplace=False)

            model_onnx_debug = oinf.run2onnx({'X': X[:3].astype(numpy.float32)})
            oinf_debug = OnnxInference(model_onnx_debug[1])

            print("DOT-SECTION", oinf_debug.to_dot())

        .. versionadded:: 0.7
        """
        intermediate = self.run(inputs, verbose=verbose, fLOG=fLOG,
                                intermediate=True, attributes=attributes)
        for name in self.input_names:
            del intermediate[name]
        new_onx = insert_results_into_onnx(
            self.obj, intermediate, as_parameter=as_parameter,
            suffix=suffix, param_name=param_name, node_type=node_type,
            domain=domain, domain_opset=domain_opset)
        return intermediate, new_onx

    def display_sequence(self, verbose=1):
        """
        Shows the sequence of nodes to run if ``runtime=='python'``.
        """
        rows = []
        rows.append(f"#node: {len(self.sequence_)}")
        for i, node in enumerate(self.sequence_):
            if verbose >= 1:
                rows.append(f"{i}: {str(node)}")
        return "\n".join(rows)

    def _run_sequence_runtime(self, inputs, clean_right_away=False,
                              intermediate=False, verbose=0, node_time=False,
                              overwrite_types=None, yield_ops=None,
                              fLOG=None, context=None, attributes=None):
        if overwrite_types is not None:
            raise NotImplementedError(  # pragma: no cover
                "overwrite_types != None not implemented.")
        if clean_right_away:
            raise NotImplementedError(  # pragma: no cover
                "clean_right_away=true not implemented.")

        if node_time:
            mtime = []
        if verbose != 0:
            printed = set()

        if context is not None:
            for k in context:
                self.global_index(k)

        if hasattr(self, "_values_init"):
            values = self._values_init.copy()  # pylint: disable=E0203
            if context is not None:
                for k, v in context.items():
                    values[self._global_index[k]] = v
        else:
            values = [None] * len(self._global_index)
            if verbose >= 1 and fLOG is not None:
                if context is not None:
                    for k, v in context.items():
                        if v is None:
                            continue
                        values[self._global_index[k]] = v
                        if verbose < 3:
                            fLOG(  # pragma: no cover
                                "+kI='{}': {} (dtype={} min={} max={})".format(
                                    k, v.shape, v.dtype, numpy_min(v), numpy_max(v)))
                        else:
                            fLOG(  # pragma: no cover
                                "+kI='{}': {} (dtype={} min={} max={}\n{}".format(
                                    k, v.shape, v.dtype, numpy_min(v), numpy_max(v), v))
                for k, v in self.inits_.items():
                    values[self._global_index[k]] = v['value']
                    if verbose < 3:
                        fLOG("+ki='{}': {} (dtype={} min={} max={})".format(
                             k, v['value'].shape, v['value'].dtype,
                             numpy_min(v['value']), numpy_max(v['value'])))
                    else:
                        fLOG("+ki='{}': {} (dtype={} min={} max={}\n{}".format(
                             k, v['value'].shape, v['value'].dtype,
                             numpy_min(v['value']), numpy_max(v['value']),
                             v['value']))
                    printed.add(k)
            else:
                if context is not None:
                    for k, v in context.items():
                        values[self._global_index[k]] = v
                for k, v in self.inits_.items():
                    values[self._global_index[k]] = v['value']
            # stores the array to skip initialing a second time
            if verbose == 0 or fLOG is None:
                self._values_init = values.copy()

        for name, value in inputs.items():
            values[self._global_index[name]] = value

        if verbose == 0 or fLOG is None:
            if node_time:
                for i, node in enumerate(self.sequence_):
                    if yield_ops is not None and node.onnx_node.op_type == 'YieldOp':
                        out = node.onnx_node.output[0]
                        if out in yield_ops:
                            values[out] = yield_ops[out]
                            continue
                        raise RuntimeError(  # pragma: no cover
                            "YieldOp output %r could not be found in "
                            "yield_ops: %r (node=%r)." % (
                                out, list(sorted(yield_ops)), node.onnx_node))
                    t = perf_counter()
                    node.run(values, attributes=attributes)
                    t2 = perf_counter()
                    mtime.append(dict(i=i, name=node.onnx_node.name,
                                      op_type=node.onnx_node.op_type,
                                      time=t2 - t))
            else:
                for node in self.sequence_:
                    node.run(values, attributes=attributes)
        else:
            def dispsimple(arr):
                if hasattr(arr, 'shape'):
                    if len(arr.shape) <= 1:
                        threshold = 8
                    else:
                        threshold = min(
                            50, min(50 // max(arr.shape[1], 1), 8) * arr.shape[1])
                    if hasattr(arr, 'todense'):
                        fLOG(  # pragma: no cover
                            numpy.array2string(arr.todense(), max_line_width=120,
                                               suppress_small=True, threshold=threshold))
                    else:
                        fLOG(numpy.array2string(arr, max_line_width=120,
                                                suppress_small=True,
                                                threshold=threshold))
                else:  # pragma: no cover
                    s = str(arr)
                    if len(s) > 50:
                        s = s[:50] + "..."
                    fLOG(s)

            if verbose >= 2:
                for k in sorted(self._global_index):
                    if values[self._global_index[k]] is None:
                        continue
                    obj = values[self._global_index[k]]
                    if k not in printed:
                        printed.add(k)
                        if hasattr(obj, 'shape'):
                            fLOG("-kv='{}' shape={} dtype={} min={} max={}{}".format(
                                k, obj.shape, obj.dtype, numpy_min(obj),
                                numpy_max(obj),
                                ' (sparse)' if isinstance(obj, coo_matrix) else ''))
                        elif (isinstance(obj, list) and len(obj) > 0 and
                                not isinstance(obj[0], dict)):  # pragma: no cover
                            fLOG(f"-kv='{k}' list len={len(obj)}")
                            if verbose >= 3 and len(obj) > 0:
                                fLOG(f"first={obj[0]} last={obj[-1]}")
                        else:  # pragma: no cover
                            fLOG(f"-kv='{k}' type={type(obj)}")

            keys = set(k for k in range(len(values)) if values[k] is not None)
            if verbose >= 1:
                fLOG("-- OnnxInference: run {} nodes with {} inputs".format(
                    len(self.sequence_), len(inputs)))
            for i, node in enumerate(self.sequence_):
                if verbose >= 1:
                    fLOG(node)
                if yield_ops is not None and node.onnx_node.op_type == 'YieldOp':
                    out = node.onnx_node.output[0]
                    if out in yield_ops:
                        fLOG(f"+yo={out!r}")
                        values[node.outputs_indices[0]] = yield_ops[out]
                    else:
                        raise RuntimeError(  # pragma: no cover
                            "YieldOp output %r could not be found in "
                            "yield_ops: %r (node=%r)." % (
                                out, list(sorted(yield_ops)), node.onnx_node))
                elif node_time:
                    t = perf_counter()
                    node.run(values, attributes=attributes)
                    t2 = perf_counter()
                    mtime.append(dict(i=i, name=node.onnx_node.name,
                                      op_type=node.onnx_node.op_type,
                                      time=t2 - t))
                else:
                    node.run(values, verbose=verbose, fLOG=fLOG,
                             attributes=attributes)
                added = 0
                for k in range(len(values)):  # pylint: disable=C0200
                    if values[k] is None:
                        continue
                    if k not in keys and k not in printed:
                        added += 1
                        printed.add(k)
                        name = list(
                            name for name in self._global_index  # pylint: disable=C0206
                            if self._global_index[name] == k)
                        if verbose >= 1:
                            if isinstance(values[k], (numpy.ndarray, coo_matrix)):
                                name = name[0]
                                mini = numpy_min(values[k])
                                maxi = numpy_max(values[k])
                                fLOG("+kr{}'{}': {} (dtype={} min={} max={}{})".format(
                                    "=" if len(values[k].shape) == 0 or min(
                                        values[k].shape) > 0 else "*",
                                    name, values[k].shape, values[k].dtype,
                                    mini, maxi,
                                    ' sparse' if isinstance(values[k], coo_matrix) else ''))
                                if verbose >= 3:
                                    dispsimple(values[k])
                            else:
                                fLOG(f"+kr='{name}': {type(values[k])}")
                                if verbose >= 3:  # pragma: no cover
                                    dispsimple(values[k])
                if added == 0 and verbose >= 1:
                    fLOG("? no new result")  # pragma: no cover

        if intermediate:
            values = [(v, k, values[v]) for k, v in self._global_index.items()]
            values.sort()
            values = OrderedDict((k, v) for _, k, v in values)
            return (values, mtime) if node_time else values

        try:
            res = {k: values[self._global_index[k]] for k in self.outputs_}
        except KeyError as e:  # pragma: no cover
            raise RuntimeError("Unable to find one output [{}]\n in [{}]"
                               ".".format(", ".join(sorted(self.outputs_)),
                                          ", ".join(sorted(values)))) from e
        if verbose != 0:
            # check input and output have the expected type
            self._validate_outputs(res, verbose=verbose, fLOG=fLOG)
        return (res, mtime) if node_time else res

    def _validate_outputs(self, res, verbose=0, fLOG=None):
        """
        Checks the output have the expected type.
        The function returns the list of mismatches.

        :param res: results in a dictionary
        :param verbose: verbosity
        :param fLOG: logging function
        :return: dictionary
        """
        if verbose >= 2:
            fLOG(f'[VALIDATE] type {type(self.obj)!r}')
        if isinstance(self.obj, ModelProto):
            from mlprodict.onnx_tools.onnx2py_helper import (
                guess_proto_dtype, get_tensor_elem_type, get_tensor_shape)
            outputs = {o.name: o for o in self.obj.graph.output}
            rows = []
            mis = {}
            for k, v in res.items():
                if k not in outputs:
                    rows.append(
                        f"Result {k!r} cannot be found in {set(outputs)!r}.")
                    continue
                try:
                    expected = get_tensor_elem_type(outputs[k])
                except TypeError:
                    expected = None
                shape = get_tensor_shape(outputs[k])
                if v is None:
                    rows.append(
                        f"Result {k!r} is None instead of {expected!r}.")
                    continue
                dtype = guess_proto_dtype(v.dtype)
                if expected != dtype:
                    mis[k] = f"dtype {dtype!r} != {expected!r}"
                    rows.append(
                        "Result %r have unexpected element type %r "
                        "instead of %r." % (
                            k, dtype, expected))
                if shape is None or len(shape) == 0:
                    continue
                if len(shape) != len(v.shape):
                    mis[k] = f"shape {v.shape!r} != {shape!r}"
                    rows.append(
                        "Result %r have unexpected shape length %r "
                        "instead of %r." % (
                            k, v.shape, shape))
                    continue
                for a, b in zip(v.shape, shape):
                    if b is None or isinstance(b, str):
                        continue
                    if a != b:
                        mis[k] = f"shape {v.shape!r} != {shape!r}"
                        rows.append(
                            "Result %r have unexpected shape %r "
                            "instead of %r." % (
                                k, v.shape, shape))
                        break
            if len(rows) > 0:
                if verbose < 0:
                    raise RuntimeError(  # pragma: no cover
                        "Validation failed.\n- %s" % "\n- ".join(rows))
                else:
                    fLOG("[VALIDATE] validation failed.\n- %s" %
                         "\n- ".join(rows))
            if verbose >= 2:  # pragma: no cover
                fLOG(f'[VALIDATE] mis={mis!r}')
            return mis

        if isinstance(self.obj, FunctionProto):
            outputs = set(self.obj.output)
            got = set(res)
            if got != outputs:
                if verbose < 0:  # pragma: no cover
                    raise RuntimeError(
                        "Unexpected mismatch between outputs %r and "
                        "expected outputs %r." % (got, outputs))
                else:  # pragma: no cover
                    fLOG(
                        f"CHECK: expected outputs {outputs!r} != outputs {got!r}")
                mis = {k: None for k in set(got) - got & outputs}
                if verbose >= 2:
                    fLOG(f'[VALIDATE] mis={mis!r}')
                return mis
            if verbose >= 2:
                fLOG('[VALIDATE] mis={}')
            return {}

        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(self.obj)!r} for self.obj.")

    def build_intermediate(self, outputs=None, verbose=0, overwrite_types=None,
                           fLOG=None):
        """
        Builds every possible :epkg:`ONNX` file
        which computes one specific intermediate output
        from the inputs.

        :param outputs: subsets of outputs to get,
            None to get all outputs,
        :param overwrite_types: shape inference does not work all the time,
            this allows to force types when building intermediate
            results, see @see fn select_model_inputs_outputs
        :param verbose: displays intermediate information
        :param fLOG: logging function
        :return: :epkg:`*py:collections:OrderedDict`

        .. versionchanged: 0.6
        """
        if verbose > 0:
            fLOG('[build_intermediate] BEGIN.')  # pragma: no cover
        if outputs is not None:
            if isinstance(outputs, str):
                outputs = [outputs]
            if not isinstance(outputs, set):
                outputs = set(outputs)
        ord = OrderedDict()
        for output in enumerate_model_node_outputs(self.obj, order=False):
            if outputs is not None and output not in outputs:
                continue
            subonx = select_model_inputs_outputs(
                self.obj, outputs=output, infer_shapes=True,
                overwrite=overwrite_types)
            subonx = onnx_remove_node_unused(subonx)
            if verbose > 0:
                fLOG(  # pragma: no cover
                    f'[build_intermediate] + {output}')
            ord[output] = OnnxInference(subonx, runtime=self.runtime,
                                        skip_run=self.skip_run,
                                        runtime_options=self.runtime_options,
                                        inplace=self.inplace,
                                        input_inplace=self.input_inplace)
        if verbose > 0:
            fLOG(  # pragma: no cover
                '[build_intermediate] END.')
        return ord

    def _run_whole_runtime(self, inputs, clean_right_away=False,
                           intermediate=False, verbose=0, node_time=False,
                           overwrite_types=None, yield_ops=None, fLOG=None,
                           context=None, attributes=None):
        # node_time is unused, context is unused
        if clean_right_away:
            raise RuntimeError(  # pragma: no cover
                "clean_right_away=true does not work with this runtime.")
        if intermediate:
            if hasattr(self, "intermediate_onnx_inference_"):
                inter_run = self.intermediate_onnx_inference_  # pylint: disable=E0203
            else:
                if verbose > 0:
                    fLOG(  # pragma: no cover
                        "-- OnnxInference: build intermediate")
                inter_run = self.build_intermediate(
                    verbose=verbose, fLOG=fLOG, overwrite_types=overwrite_types)
                self.intermediate_onnx_inference_ = inter_run
                graph = self.to_sequence()
                self.inits_ = graph['inits']

            if verbose >= 1:
                fLOG(  # pragma: no cover
                    "-- OnnxInference: run {} nodes".format(
                        len(self.intermediate_onnx_inference_)))
            values = OrderedDict(inputs)
            for k, v in self.inits_.items():
                values[k] = v['value']
            if verbose >= 2:  # pragma: no cover
                for k in sorted(values):
                    fLOG(
                        f"-k='{k}' shape={values[k].shape} dtype={values[k].dtype}")
            for node, oinf in self.intermediate_onnx_inference_.items():
                if verbose >= 4:  # pragma: no cover
                    fLOG(f'[intermediate] {node!r}')
                    if verbose >= 5:  # pragma: no cover
                        fLOG(oinf.obj)
                if yield_ops is not None and node.onnx_node.op_type == 'YieldOp':
                    out = node.onnx_node.output[0]
                    if out in yield_ops:
                        values[out] = yield_ops[out]
                        continue
                    raise RuntimeError(  # pragma: no cover
                        "YieldOp output %r could not be found in "
                        "yield_ops: %r (node=%r)." % (
                            out, list(sorted(yield_ops)), node.onnx_node))
                output = oinf.run(inputs, attributes=attributes)[node]
                values[node] = output
                if verbose >= 1:
                    if verbose >= 4:  # pragma: no cover
                        for k, v in inputs.items():
                            if isinstance(output, numpy.ndarray):
                                fLOG("-i='{}': {} (dtype={}) {}".format(
                                     k, v.shape, v.dtype, v.ravel().tolist()))
                            else:
                                fLOG(
                                    f"-i='{k}': {v.shape} (dtype={v.dtype}) - ?")
                    if isinstance(output, numpy.ndarray):
                        fLOG("+k='{}': {} (dtype={})".format(  # pragma: no cover
                            node, output.shape, output.dtype))
                        if verbose >= 2:  # pragma: no cover
                            fLOG(output)
                    else:
                        fLOG("+k='{}': {}".format(  # pragma: no cover
                            node, type(output)))
                        if verbose >= 2:  # pragma: no cover
                            fLOG(output)
            return values

        if verbose != 0:
            warnings.warn(
                "verbose option not implemented if runtime is 'onnxruntime1'")
        res = self._whole.run(inputs)
        return {k: v for k, v in zip(self.outputs_, res)}

    def __getitem__(self, item):
        """
        Returns the ONNX verions of a node.
        """
        if isinstance(item, tuple):
            node_name, att_name = item
        else:
            node_name = item
            att_name = None

        node_ = None
        for node in self.obj.graph.node:
            if node.name == node_name:
                node_ = node
                break

        if node_ is None:
            raise IndexError(  # pragma: no cover
                "Unable to get node name '{}'.\n{}".format(
                    node_name, "\n".join(node.name for node in self.obj.graph.node)))

        if att_name is None:
            return node_

        for att in node_.attribute:
            if att.name == att_name:
                return att

        raise IndexError(  # pragma: no cover
            f"Unable to find attribute '{att_name}' from node '{node_name}'.")

    def switch_initializers_dtype(self, model=None,
                                  dtype_in=numpy.float32,
                                  dtype_out=numpy.float64):
        """
        Switches all initializers to ``numpy.float64``. If *model*
        is None, a simple cast is done. Otherwise, the function assumes
        the model is a :epkg:`scikit-learn` pipeline.
        This only works if the runtime is ``'python'``.

        @param      model       :epkg:`scikit-learn` model or None
        @param      dtype_in    previous type
        @param      dtype_out   next type
        @return                 done operations
        """
        from ..onnx_tools.optim.sklearn_helper import enumerate_fitted_arrays, pairwise_array_distances

        if self.runtime != 'python':  # pragma: no cover
            raise RuntimeError("Initializers can be casted only if the "
                               "runtime is 'python' not '{}'.".format(self.runtime))

        if hasattr(self, '_values_init'):
            del self._values_init

        # first pass: simple cast
        done = []
        initializer = self.inits_
        for k, v in initializer.items():
            if isinstance(v['value'], numpy.ndarray):
                if v['value'].dtype == dtype_in:
                    v['value'] = v['value'].astype(dtype_out)
                    done.append(("pass1", "+", "init", k, v['value']))
                else:
                    done.append(("pass1", "-", "init", k,
                                 v['value']))  # pragma: no cover
        for k, v in self.graph_['nodes'].items():
            res = v.switch_initializers_dtype(dtype_in=dtype_in,
                                              dtype_out=dtype_out)
            for r in res:
                done.append(("pass1", "node", k) + r)
        for k, v in self.graph_['intermediate'].items():
            if v is None:
                continue
            res = v.switch_initializers_dtype(dtype_in=dtype_in,
                                              dtype_out=dtype_out)
            for r in res:
                done.append(("pass1", "sub", k) + r)

        if model is not None:
            # Second pass, we compare all arrays from the model
            # to the arrays in the converted models.
            def dist(a):
                cast = a.astype(dtype_in).astype(dtype_out)
                d = pairwise_array_distances([cast], [a])[0, 0]
                return d

            done_ = [(c, c[-1]) for c in done]
            moda_ = [(a, a[-2][-1]) for a in enumerate_fitted_arrays(model)
                     if dist(a[-2][-1]) > 0]
            aconv = [_[-1] for _ in done_]
            amoda = [_[-1] for _ in moda_]
            distances = pairwise_array_distances(aconv, amoda)

            for i in range(distances.shape[0]):
                j = numpy.argmin(distances[i])
                d = distances[i, j]
                if d < 0.1:
                    numpy.copyto(aconv[i], amoda[j])
                    done.append(("pass2", d) + done_[i][0])

        return done

    def _set_shape_inference_runtime(self):
        """
        Set shapes based on shape inference
        relying on the runtime.
        The values are stored in every node.
        """
        try:
            rt = OnnxShapeInference(self.obj)
        except (ShapeInferenceMissing, NotImplementedShapeInferenceError,
                ShapeInferenceDimensionError, NotImplementedError):
            # an operator is missing, shape cannot be computed.
            return {name: None for name in self.output_names}
        except KeyError:
            # subgraphs or functions are not yet handled.
            # it should be removed later.
            return {name: None for name in self.output_names}
        except NameError:
            # loop subgraphs or function are not yet handled.
            # they may overwrite results.
            return {name: None for name in self.output_names}
        except (ShapeInferenceException, RuntimeError, IndexError) as e:
            raise ShapeInferenceException(  # pragma: no cover
                f"Unable to run ShapeInference for\n{str(self.obj)}") from e
        out = rt.run()
        values = out.get()
        return values

    def infer_shapes(self):
        """
        Computes expected shapes.

        :return: dictionary of shapes
        """
        return self._set_shape_inference_runtime()

    def _guess_inplace(self, input_inplace=False):
        """
        Looks into every node of the graph to see
        if there is a way to do the computation
        inplace. By default (*input_inplace=False*),
        the function assumes inputs cannot be modified
        so the first node cannot do inplace computation.
        This function only works with the python runtime.

        @param      input_inplace       the computation is allowed
                                        to overwrite the input

        This function checks that one node is used only
        once and then can be modified by the next node.
        Nodes `A`, `C` can be overwritten by the computation.
        Node `B` cannot as it is used by two nodes.

        .. blockdiag::

            diagram {
                A -> B -> C -> E;
                     B -> D;
            }

        It does not handle specific case such node `B` being
        overwritten by node `C` but without changing its shape
        and node `D` only needs the shape of `B`. Then `B` could
        be overwritten as well.
        """
        forbid = {}
        values = OrderedDict()
        for k in self.statics_:
            values[k] = dict(inplace=False, to=[], fr=[])
        for k in self.inputs_:
            values[k] = dict(inplace=input_inplace, to=[], fr=[])
        for k in self.inits_:
            values[k] = dict(inplace=False, to=[], fr=[])
        for node in self.sequence_:
            for n in node.inputs:
                if n == '':
                    continue
                values[n]['to'].append(node)
            for n in node.outputs:
                if node.op_type == 'Constant':
                    # We cannot modify constant.
                    forbid[n] = node
                if n not in values:
                    values[n] = dict(inplace=None, to=[], fr=[])
                values[n]['fr'].append(node)

        # checks the number of outputs
        outputs = set(self.output_names)
        modif = 1
        while modif > 0:
            modif = 0
            for n, v in values.items():
                if v['inplace'] is not None:
                    continue
                if n in forbid:
                    continue
                if len(v['to']) == 1:
                    v['inplace'] = True
                    modif += 1

        # convey the information to every node
        inplaces = {}
        for n, v in values.items():
            if v['inplace']:
                inplaces[n] = v
                for node in v['to']:
                    if n in outputs:
                        continue
                    node.enable_inplace_compute(n)

        return inplaces

    def _build_compile_run(self, debug=False):
        """
        Rewrite the run function in python,
        compiles it, and adds it as a method.

        @param      debug       insert debugging code
        @return                 method name, callable object

        .. exref::
            :title: Run a model with runtime 'python_compiled'

            The following code trains a model and compute
            the predictions with runtime ``'python_compiled'``.
            It converts the onnx graph into a python function
            which calls every operator. Its code is printed
            below.

            .. runpython::
                :showcode:
                :warningout: DeprecationWarning

                import numpy
                from sklearn.datasets import load_iris
                from sklearn.model_selection import train_test_split
                from sklearn.ensemble import AdaBoostClassifier
                from sklearn.tree import DecisionTreeClassifier
                from mlprodict.onnx_conv import to_onnx
                from mlprodict.onnxrt import OnnxInference

                iris = load_iris()
                X, y = iris.data, iris.target
                X_train, X_test, y_train, __ = train_test_split(X, y, random_state=11)
                y_train = y_train.astype(numpy.float32)
                clr = AdaBoostClassifier(
                    base_estimator=DecisionTreeClassifier(max_depth=3),
                    n_estimators=3)
                clr.fit(X_train, y_train)

                model_def = to_onnx(clr, X_train.astype(numpy.float32),
                                    target_opset=12)

                oinf2 = OnnxInference(model_def, runtime='python_compiled')
                print(oinf2.run({'X': X_test[:5]}))

                # prints out the python function equivalent
                # to the onnx graph
                print(oinf2)
        """

        def clean_name(name):
            res = name.replace(":", "_").replace('.', '_').replace('/', '_')
            if iskeyword(res):
                res += '_'
            return res

        # inits
        inputs = self.input_names
        code = [
            'def compiled_run(dict_inputs, yield_ops=None, context=None, attributes=None):']
        code.append("    if yield_ops is not None:")
        code.append("        raise NotImplementedError"
                    "('yields_ops should be None.')")
        if debug:
            code.append("    printed = {}")

        context = {}

        # static variables
        for k in sorted(self.statics_):
            code.append(f"    # static: {k}")
            code.append(f"    {clean_name(k)} = dict_inputs['{k}']")
            if debug:  # pragma: no cover
                code.append(
                    f"    debug_print('i.{clean_name(k)}', {k}, printed)")

        # initializers
        for k, v in sorted(self.inits_.items()):
            if k.startswith("_OPT_"):
                raise RuntimeError(  # pragma: no cover
                    "The runtime cannot handle any constant name "
                    "starting with '_OPT_': '{}'.".format(k))
            if k in inputs:
                context["_OPT_" + clean_name(k)] = v['value']
                code.append(f"    # init: _OPT_{clean_name(k)} ({k})")
                if debug:  # pragma: no cover
                    code.append(
                        "    debug_print('c.[_OPT_{0}]', _OPT_{1}, printed)".format(
                            clean_name(k), k))
            else:
                context[clean_name(k)] = v['value']
                code.append(f"    # init: {clean_name(k)} ({k})")
                if debug:
                    code.append(
                        f"    debug_print('c.[{clean_name(k)}]', {k}, printed)")

        # method signature
        code.append("    # inputs")
        for inp in inputs:
            if '_OPT_' + inp in context:
                # optional inputs
                code.append(
                    "    {0} = dict_inputs.get('{1}', _OPT_{0})".format(
                        clean_name(inp), inp))
            else:
                code.append(f"    {clean_name(inp)} = dict_inputs['{inp}']")
            if debug:
                code.append(
                    f"    debug_print('i.{clean_name(inp)}', {inp}, printed)")

        # code
        for i, node in enumerate(self.sequence_):
            name = f"n{i}_{node.ops_.__class__.__name__.lower()}"
            if node.ops_ is None:
                context[name] = node.function_
                # The code of the function should be added but only once.
                raise NotImplementedError(
                    "Not implemented for models including functions.")
            else:
                context[name] = node.ops_._run
                if (node.ops_.__class__.__name__ == 'Loop' and
                        node.ops_.need_context()):
                    # Adding context.
                    ctx = "{%s}" % ", ".join(
                        "'%s': %s" % (n, n) for n in node.ops_.additional_inputs)
                    code.append('    ({1}, ) = {2}({0}, context={3})'.format(
                        ', '.join(map(clean_name, node.inputs)),
                        ', '.join(map(clean_name, node.outputs)),
                        name, ctx))
                else:
                    code.append('    ({1}, ) = {2}({0})'.format(
                        ', '.join(map(clean_name, node.inputs)),
                        ', '.join(map(clean_name, node.outputs)),
                        name))
                if debug:
                    code.append(f"    print('''# {code[-1][4:]}''')")
                    for o in node.outputs:
                        code.append(
                            f"    debug_print('o.{clean_name(o)}', {o}, printed)")

        # return
        code.append('    return {')
        for out in self.output_names:
            code.append(f"        '{out}': {clean_name(out)},")
        code.append('    }')
        final_code = '\n'.join(code)

        # compile the outcome
        context['self'] = self
        try:
            obj = compile(final_code, "<string>", 'exec')
        except SyntaxError as e:  # pragma: no cover
            raise SyntaxError(
                f"Unable to compile\n#####\n{final_code}") from e
        fcts_obj = [_ for _ in obj.co_consts
                    if _ is not None and not isinstance(_, (bool, str, int))]
        fct = make_callable(
            "compiled_run", fcts_obj[0], final_code, context, debug)

        # end
        return "compiled_run", fct, final_code

    def reduce_size(self, pickable=False):
        """
        Reduces the memory footprint as much as possible.

        @param  pickable        keeps a pickle object?
        """
        import gc
        del self.graph_
        if not pickable:
            del self.obj
        if self.runtime in ('python_compiled', 'python_compiled_debug'):
            del self.sequence_
        gc.collect()

    def get_profiling(self, as_df=False):
        """
        Returns the profiling after a couple of execution.

        :param as_df: return the results as a dataframe (True)
        :return: dataframe or list of dictionaries

        .. versionadded:: 0.6
        """
        if (self.runtime_options is None or
                not self.runtime_options.get('enable_profiling', False)):
            raise RuntimeError(
                "Profiling is available if options 'enable_profiling' "
                "is set to true in 'runtime_options' but is %r." % self.runtime_options)
        prof = None
        if hasattr(self, '_whole'):
            prof = self._whole.get_profiling()
        if prof is None:
            raise NotImplementedError(  # pragma: no cover
                "profiling is only implemented for runtime 'onnxruntime1'.")
        if as_df:
            import pandas
            return pandas.DataFrame(prof)
        return prof

    def get_execution_order(self):
        """
        This function returns a dictionary `{(kind, name): (order, op)}`,
        *name* can be a node name or a result name. In that case,
        it gets the execution order than the node which created it.
        The function returns None if the order is not available
        (the selected runtime does not return it). *kind* is either
        `'node'` or `'node'`. If two nodes have the same name,
        returned order is the last one. Initializers gets an execution
        order equal to -1, inputs to 0, all others results are >= 1.

        .. versionadded:: 0.7
        """
        if not hasattr(self, "sequence_"):
            return None

        res = {}
        for k, v in self.inits_.items():
            res['res', k] = (-1, v)
        for name, shape in self.input_names_shapes:
            res['res', name] = (0, shape)

        for i, node in enumerate(self.sequence_):
            key = ('node', node.onnx_node.name)
            res[key] = (i + 1, node)
            for out in node.onnx_node.output:
                key = ('res', out)
                if key in res:
                    raise RuntimeError(  # pragma: no cover
                        "Output %r of node name %r already registered."
                        "" % (out, node.onnx_node.name))
                res[key] = (i + 1, None)

        return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Binarizer(OpRunUnaryNum):

    atts = {'threshold': 0.}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Binarizer.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        X = x.copy()
        cond = X > self.threshold
        not_cond = numpy.logical_not(cond)
        X[cond] = 1
        X[not_cond] = 0
        return (X, )



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *npy*.

.. versionadded:: 0.6
"""
from .onnx_numpy_annotation import (
    NDArray, NDArrayType, NDArraySameType, NDArraySameTypeSameShape,
    Shape, DType)
from .onnx_numpy_compiler import OnnxNumpyCompiler
from .onnx_numpy_wrapper import onnxnumpy, onnxnumpy_default, onnxnumpy_np
from .onnx_sklearn_wrapper import (
    update_registered_converter_npy, onnxsklearn_class,
    onnxsklearn_transformer, onnxsklearn_regressor,
    onnxsklearn_classifier, onnxsklearn_cluster)
from .onnx_version import FctVersion



"""
@file
@brief Functions which converts :epkg:`ONNX` object into
readable :epkg:`python` objects.
"""
import pprint
import warnings
import numpy
from scipy.sparse import coo_matrix
from onnx.defs import get_schema, get_function_ops, onnx_opset_version
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE, TENSOR_TYPE_TO_NP_TYPE
from onnx import TensorProto, ValueInfoProto, TypeProto
from onnx.helper import make_tensor_type_proto
from onnx.numpy_helper import to_array, from_array as onnx_from_array


def get_tensor_shape(obj):
    """
    Returns the shape if that makes sense for this object.
    """
    if isinstance(obj, ValueInfoProto):
        return get_tensor_shape(obj.type)
    elif not isinstance(obj, TypeProto):
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(obj)!r}.")
    if not obj.tensor_type.HasField('shape'):
        return None
    shape = []
    for d in obj.tensor_type.shape.dim:
        v = d.dim_value if d.dim_value > 0 else d.dim_param
        shape.append(v)
    if len(shape) == 0:
        return shape
    return list(None if s in (0, '') else s for s in shape)


def get_tensor_elem_type(obj):
    """
    Returns the element type if that makes sense for this object.
    """
    if isinstance(obj, ValueInfoProto):
        return get_tensor_elem_type(obj.type)
    elif not isinstance(obj, TypeProto):
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(obj)!r}.")
    if obj.tensor_type.ByteSize() == 0:
        raise TypeError(  # pragma: no cover
            f"Unable to guess element type for {obj!r}.")
    return obj.tensor_type.elem_type


def to_bytes(val):
    """
    Converts an array into protobuf and then into bytes.

   :param val: array
   :return: bytes

    .. exref::
        :title: Converts an array into bytes (serialization)

        Useful to serialize.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning

            import numpy
            from mlprodict.onnx_tools.onnx2py_helper import to_bytes

            data = numpy.array([[0, 1], [2, 3], [4, 5]], dtype=numpy.float32)
            pb = to_bytes(data)
            print(len(pb), data.size * data.itemsize, pb[:10])
    """
    if isinstance(val, numpy.ndarray):
        pb = from_array(val)
    else:
        pb = val  # pragma: no cover
    return pb.SerializeToString()


def from_array(value, name=None):
    """
    Converts an array into an ONNX tensor.

    :param value: numpy array
    :return: ONNX tensor
    """
    if isinstance(value, numpy.ndarray):
        try:
            pb = onnx_from_array(value, name=name)
        except NotImplementedError as e:  # pragma: no cover
            if value.dtype == numpy.dtype('O'):
                pb = TensorProto()
                pb.data_type = TensorProto.STRING  # pylint: disable=E1101
                if name is not None:
                    pb.name = name
                pb.dims.extend(value.shape)  # pylint: disable=E1101
                pb.string_data.extend(  # pylint: disable=E1101
                    list(map(lambda o: str(o).encode('utf-8'), value.ravel())))
            else:
                raise NotImplementedError(
                    "Unable to convert type %r (dtype=%r) into an ONNX tensor "
                    "due to %r." % (type(value), value.dtype, e)) from e
        return pb
    if isinstance(value, TensorProto):  # pragma: no cover
        return value
    raise NotImplementedError(  # pragma: no cover
        f"Unable to convert type {type(value)!r} into an ONNX tensor.")


def from_bytes(b):
    """
    Retrieves an array from bytes then protobuf.

    :param b: bytes
    :return: array

    .. exref::
        :title: Converts bytes into an array (serialization)

        Useful to deserialize.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning

            import numpy
            from mlprodict.onnx_tools.onnx2py_helper import to_bytes, from_bytes

            data = numpy.array([[0, 1], [2, 3], [4, 5]], dtype=numpy.float32)
            pb = to_bytes(data)
            data2 = from_bytes(pb)
            print(data2)
    """
    if isinstance(b, bytes):
        pb = TensorProto()
        pb.ParseFromString(b)
    else:
        pb = b  # pragma: no cover
    return to_array(pb)


def _numpy_array(data, dtype=None, copy=True):
    """
    Single function to create an array.

    @param      data        data
    @param      dtype       dtype
    @param      copy        copy
    @return                 numpy array
    """
    if isinstance(data, numpy.ndarray):
        res = data
    else:
        res = numpy.array(data, dtype=dtype, copy=copy)
    return res


def _sparse_array(shape, data, indices, dtype=None, copy=True):
    """
    Single function to create an sparse array
    (:epkg:`coo_matrix`).

    @param      shape       shape
    @param      data        data
    @param      indices     indices
    @param      dtype       dtype
    @param      copy        copy
    @return                 :epkg:`coo_matrix`
    """
    if len(shape) != 2:
        raise ValueError(  # pragma: no cover
            f"Only matrices are allowed or sparse matrices but shape is {shape}.")
    rows = numpy.array([i // shape[1] for i in indices])
    cols = numpy.array([i % shape[1] for i in indices])
    if isinstance(data, numpy.ndarray):
        res = coo_matrix((data, (rows, cols)), dtype=dtype)
    else:
        res = coo_matrix(  # pragma: no cover
            (numpy.array(data, dtype=dtype, copy=copy),
             (rows, cols)), dtype=dtype)
    return res


def guess_numpy_type_from_string(name):
    """
    Converts a string (such as `'float'`) into a
    numpy dtype.
    """
    if name in ('float', 'float32'):
        return numpy.float32
    if name in ('double', 'float64'):
        return numpy.float64
    if name == 'float16':
        return numpy.float16
    if name == 'int64':
        return numpy.int64
    if name == 'int8':
        return numpy.int8
    if name == 'uint8':
        return numpy.uint8
    if name == 'int32':
        return numpy.int32
    if name == 'int16':
        return numpy.int16
    if name == 'bool':
        return numpy.bool_
    if name == 'str':
        return numpy.str_
    raise ValueError(  # pragma: no cover
        f"Unable to guess numpy dtype from {name!r}.")


def guess_numpy_type_from_dtype(dt):
    """
    Converts a string (such as `'dtype(float32)'`) into a
    numpy dtype.
    """
    if dt in {numpy.int8, numpy.uint8, numpy.float16, numpy.float32,
              numpy.float64, numpy.int32, numpy.int64, numpy.int16,
              numpy.uint16, numpy.uint32, numpy.bool_, numpy.str_,
              numpy.uint64, bool, str, }:
        return dt
    if dt == numpy.dtype('float32'):
        return numpy.float32
    if dt == numpy.dtype('float64'):
        return numpy.float64
    if dt == numpy.dtype('int64'):
        return numpy.int64
    if dt == numpy.dtype('int8'):
        return numpy.int8
    if dt == numpy.dtype('uint8'):
        return numpy.uint8
    raise ValueError(  # pragma: no cover
        f"Unable to guess numpy dtype from {dt!r}.")


def _elem_type_as_str(elem_type):
    if elem_type == TensorProto.FLOAT:  # pylint: disable=E1101
        return 'float'
    if elem_type == TensorProto.BOOL:  # pylint: disable=E1101
        return 'bool'
    if elem_type == TensorProto.DOUBLE:  # pylint: disable=E1101
        return 'double'
    if elem_type == TensorProto.STRING:  # pylint: disable=E1101
        return 'str'
    if elem_type == TensorProto.INT64:  # pylint: disable=E1101
        return 'int64'
    if elem_type == TensorProto.INT32:  # pylint: disable=E1101
        return 'int32'
    if elem_type == TensorProto.UINT32:  # pylint: disable=E1101
        return 'uint32'
    if elem_type == TensorProto.UINT64:  # pylint: disable=E1101
        return 'uint64'
    if elem_type == TensorProto.INT16:  # pylint: disable=E1101
        return 'int16'
    if elem_type == TensorProto.UINT16:  # pylint: disable=E1101
        return 'uint16'
    if elem_type == TensorProto.UINT8:  # pylint: disable=E1101
        return 'uint8'
    if elem_type == TensorProto.INT8:  # pylint: disable=E1101
        return 'int8'
    if elem_type == TensorProto.FLOAT16:  # pylint: disable=E1101
        return 'float16'
    if elem_type == TensorProto.COMPLEX64:  # pylint: disable=E1101
        return 'complex64'
    if elem_type == TensorProto.COMPLEX128:  # pylint: disable=E1101
        return 'complex128'
    if elem_type == 0:  # pylint: disable=E1101
        return 'unk'

    # The following code should be refactored.
    selem = str(elem_type)

    if selem.startswith("tensor_type"):
        this = elem_type.tensor_type
        et = _elem_type_as_str(this.elem_type)
        shape = this.shape
        dim = shape.dim
        dims = [d.dim_value for d in dim]
        if len(dims) == 0:
            dims = '?'
        return {'kind': 'tensor', 'elem': et, 'shape': shape}

    if selem.startswith("optional_type"):
        this = elem_type.optional_type
        et = _elem_type_as_str(this.elem_type)
        shape = this.shape
        dim = shape.dim
        dims = [d.dim_value for d in dim]
        if len(dims) == 0:
            dims = '?'
        return {'kind': 'tensor', 'elem': et, 'shape': shape,
                'optional_type': True}

    if selem.startswith("map_type"):
        this = elem_type.map_type
        kt = _elem_type_as_str(this.key_type)
        vt = _elem_type_as_str(this.value_type)
        return {'kind': 'map', 'key': kt, 'value': vt}

    raise NotImplementedError(  # pragma: no cover
        "elem_type '{}' is unknown\nfields:\n{}\n-----\n{}.".format(
            elem_type, pprint.pformat(dir(elem_type)), type(elem_type)))


def _to_array(var):
    try:
        data = to_array(var)
    except ValueError as e:  # pragma: no cover
        dims = [d for d in var.dims]
        if var.data_type == 1 and var.float_data is not None:
            try:
                data = _numpy_array(var.float_data, dtype=numpy.float32,
                                    copy=False).reshape(dims)
            except ValueError:
                data = _numpy_array(to_array(var))
        elif var.data_type == 2 and var.uint8_data is not None:
            data = _numpy_array(var.uint8_data, dtype=numpy.uint8,
                                copy=False).reshape(dims)
        elif var.data_type == 3 and var.int8_data is not None:
            data = _numpy_array(var.int8_data, dtype=numpy.int8,
                                copy=False).reshape(dims)
        elif var.data_type == 4 and var.uint16_data is not None:
            data = _numpy_array(var.uint16_data, dtype=numpy.uint16,
                                copy=False).reshape(dims)
        elif var.data_type == 5 and var.int16_data is not None:
            data = _numpy_array(var.int16_data, dtype=numpy.int16,
                                copy=False).reshape(dims)
        elif var.data_type == 6 and var.int32_data is not None:
            data = _numpy_array(var.int32_data, dtype=numpy.int32,
                                copy=False).reshape(dims)
        elif var.data_type == 7 and var.int64_data is not None:
            data = _numpy_array(var.int64_data, dtype=numpy.int64,
                                copy=False).reshape(dims)
        elif var.data_type == 11 and var.double_data is not None:
            try:
                data = _numpy_array(var.double_data, dtype=numpy.float64,
                                    copy=False).reshape(dims)
            except ValueError:
                data = _numpy_array(to_array(var))
        elif var.data_type == 16 and var.float16_data is not None:
            data = _numpy_array(var.float16_data, dtype=numpy.float16,
                                copy=False).reshape(dims)
        else:
            raise NotImplementedError(
                f"Iniatilizer {var} cannot be converted into a dictionary.") from e
    return data


def _var_as_dict(var):
    """
    Converts a protobuf object into something readable.
    The current implementation relies on :epkg:`json`.
    That's not the most efficient way.
    """
    if hasattr(var, 'type') and str(var.type) != '':
        # variable
        if var.type is not None:
            if hasattr(var, 'sparse_tensor') and var.type == 11:
                # sparse tensor
                t = var.sparse_tensor
                values = _var_as_dict(t.values)
                dims = list(t.dims)
                dtype = dict(kind='sparse_tensor', shape=tuple(dims), elem=1)
            elif (hasattr(var.type, 'tensor_type') and
                    var.type.tensor_type.elem_type > 0):
                t = var.type.tensor_type
                elem_type = _elem_type_as_str(t.elem_type)
                shape = t.shape
                dim = shape.dim
                dims = [d.dim_value for d in dim]
                if len(dims) == 0:
                    dims = '?'
                dtype = dict(kind='tensor', elem=elem_type,
                             shape=tuple(dims))
            elif (hasattr(var.type, 'optional_type') and
                    var.type.tensor_type.elem_type > 0):
                t = var.type.optional_type
                elem_type = _elem_type_as_str(t.elem_type)
                shape = t.shape
                dim = shape.dim
                dims = [d.dim_value for d in dim]
                if len(dims) == 0:
                    dims = '?'
                dtype = dict(kind='tensor', elem=elem_type,
                             shape=tuple(dims), optional_type=True)
            elif (hasattr(var.type, 'real') and var.type.real == 5 and
                    hasattr(var, 'g')):
                dtype = dict(kind='graph', elem=var.type.real)
            elif (hasattr(var.type, 'real') and var.type.real == 4 and
                    hasattr(var, 't')):
                dtype = dict(kind='tensor', elem=var.type.real)
            elif hasattr(var.type, 'real'):
                dtype = dict(kind='real', elem=var.type.real)
            elif (hasattr(var.type, "sequence_type") and
                    var.type.sequence_type is not None and
                    str(var.type.sequence_type.elem_type) != ''):
                t = var.type.sequence_type
                elem_type = _elem_type_as_str(t.elem_type)
                dtype = dict(kind='sequence', elem=elem_type)
            elif (hasattr(var.type, "map_type") and
                    var.type.map_type is not None and
                    str(var.type.map_type.key_type) != '' and
                    str(var.type.map_type.value_type) != ''):
                t = var.type.map_type
                key_type = _elem_type_as_str(t.key_type)
                value_type = _elem_type_as_str(t.value_type)
                dtype = dict(kind='map', key=key_type, value=value_type)
            elif (hasattr(var.type, 'tensor_type') and
                    var.type.tensor_type.elem_type == 0):
                if hasattr(var.type, 'optional_type'):
                    optional = var.type.optional_type
                else:
                    optional = None
                t = var.type.tensor_type
                elem_type = _elem_type_as_str(t.elem_type)
                shape = t.shape
                dim = shape.dim
                dims = [d.dim_value for d in dim]
                if len(dims) == 0:
                    dims = '?'
                dtype = dict(kind='tensor', elem=elem_type,
                             shape=tuple(dims))
                if optional is not None:
                    dtype['optional'] = _var_as_dict(optional)
            else:
                raise NotImplementedError(  # pragma: no cover
                    "Unable to convert a type into a dictionary for '{}'. "
                    "Available fields: {}.".format(
                        var.type, pprint.pformat(dir(var.type))))
        else:
            raise NotImplementedError(  # pragma: no cover
                "Unable to convert variable into a dictionary for '{}'. "
                "Available fields: {}.".format(
                    var, pprint.pformat(dir(var.type))))

        res = dict(name=var.name, type=dtype)

        if (hasattr(var, 'sparse_tensor') and dtype.get('elem', None) == 1 and
                dtype['kind'] == 'sparse_tensor'):
            # sparse matrix
            t = var.sparse_tensor
            try:
                values = _var_as_dict(t.values)
            except NotImplementedError as e:  # pragma: no cover
                raise NotImplementedError(
                    f"Issue with\n{var}\n---") from e
            indices = _var_as_dict(t.indices)
            res['value'] = _sparse_array(
                dtype['shape'], values['value'], indices['value'], dtype=numpy.float32)
        elif hasattr(var, 'floats') and dtype.get('elem', None) == 6:
            res['value'] = _numpy_array(var.floats, dtype=numpy.float32)
        elif hasattr(var, 'strings') and dtype.get('elem', None) == 8:
            res['value'] = _numpy_array(var.strings)
        elif hasattr(var, 'ints') and dtype.get('elem', None) == 7:
            res['value'] = _numpy_array(var.ints)
        elif hasattr(var, 'f') and dtype.get('elem', None) == 1:
            res['value'] = var.f
        elif hasattr(var, 's') and dtype.get('elem', None) == 3:
            res['value'] = var.s
        elif hasattr(var, 'i') and dtype.get('elem', None) == 2:
            res['value'] = var.i
        elif hasattr(var, 'g') and dtype.get('elem', None) == 5:
            res['value'] = var.g
        elif hasattr(var, 't') and dtype.get('elem', None) == 4:
            if hasattr(var, 'ref_attr_name') and var.ref_attr_name:
                res['ref_attr_name'] = var.ref_attr_name
            else:
                ts = _var_as_dict(var.t)
                res['value'] = ts['value']
        elif hasattr(var, 'sparse_tensor') and dtype.get('elem', None) == 11:
            ts = _var_as_dict(var.sparse_tensor)
            if hasattr(var, 'ref_attr_name') and var.ref_attr_name:
                res['ref_attr_name'] = var.ref_attr_name
            else:
                ts = _var_as_dict(var.t)
                res['value'] = ts['value']
        elif "'value'" in str(var):
            warnings.warn("No value: {} -- {}".format(  # pragma: no cover
                dtype, str(var).replace("\n", "").replace(" ", "")))
        return res

    if hasattr(var, 'op_type'):
        if hasattr(var, 'attribute'):
            atts = {}
            for att in var.attribute:
                atts[att.name] = _var_as_dict(att)
        return dict(name=var.name, op_type=var.op_type,
                    domain=var.domain, atts=atts)
    if hasattr(var, 'dims') and len(var.dims) > 0:
        # initializer
        data = _to_array(var)
        return dict(name=var.name, value=data)
    if hasattr(var, 'data_type') and var.data_type > 0:
        data = _to_array(var)
        return dict(name=var.name, value=data)
    if isinstance(var, str):
        return dict(name=var)
    if str(var) == '':
        return None
    if isinstance(var, ValueInfoProto):
        return dict(name=var.name,
                    type=dict(elem='unk', kind='tensor', shape=('?', )))
    raise NotImplementedError(  # pragma: no cover
        "Unable to guess which object it is type is %r value is %r "
        "(hasattr(var,'type')=%r, var.type=%s."
        "" % (type(var), str(var), hasattr(var, 'type'),
              str(getattr(var, 'type', None))))


def get_dtype_shape(obj):
    """
    Returns the shape of a tensor.

    :param obj: onnx object
    :return: `(dtype, shape)` or `(None, None)` if not applicable
    """
    if not hasattr(obj, 'type'):
        return None
    t = obj.type
    if not hasattr(t, 'tensor_type'):
        return None
    t = t.tensor_type
    dtype = t.elem_type
    if not hasattr(t, 'shape'):
        return dtype, None
    shape = t.shape
    ds = []
    for dim in shape.dim:
        d = dim.dim_value
        s = dim.dim_param
        if d == 0:
            if s == '':
                ds.append(None)
            else:
                ds.append(s)
        else:
            ds.append(d)
    return dtype, tuple(ds)


def onnx_model_opsets(onnx_model):
    """
    Extracts opsets in a dictionary.

    :param onnx_model: ONNX graph
    :return: dictionary `{domain: version}`
    """
    res = {}
    for oimp in onnx_model.opset_import:
        res[oimp.domain] = oimp.version
    return res


def _type_to_string(dtype):
    """
    Converts a type into a readable string.
    """
    if not isinstance(dtype, dict):
        dtype_ = _var_as_dict(dtype)  # pragma: no cover
    else:
        dtype_ = dtype
    if dtype_["kind"] == 'tensor':
        return f"{dtype_['elem']}({dtype_['shape']})"
    if dtype_['kind'] == 'sequence':
        return f"[{_type_to_string(dtype_['elem'])}]"
    if dtype_["kind"] == 'map':
        return f"{{{dtype_['key']}, {dtype_['value']}}}"
    raise NotImplementedError(  # pragma: no cover
        f"Unable to convert into string {dtype} or {dtype_}.")


def numpy_min(x):
    """
    Returns the minimum of an array.
    Deals with text as well.
    """
    try:
        if hasattr(x, 'todense'):
            x = x.todense()
        if x.dtype.kind not in 'cUC':
            return x.min()
        try:  # pragma: no cover
            x = x.ravel()
        except AttributeError:  # pragma: no cover
            pass
        keep = list(filter(lambda s: isinstance(s, str), x))
        if len(keep) == 0:  # pragma: no cover
            return numpy.nan
        keep.sort()
        val = keep[0]
        if len(val) > 10:  # pragma: no cover
            val = val[:10] + '...'
        return f"{val!r}"
    except (ValueError, TypeError):  # pragma: no cover
        return '?'


def numpy_max(x):
    """
    Returns the maximum of an array.
    Deals with text as well.
    """
    try:
        if hasattr(x, 'todense'):
            x = x.todense()
        if x.dtype.kind not in 'cUC':
            return x.max()
        try:  # pragma: no cover
            x = x.ravel()
        except AttributeError:  # pragma: no cover
            pass
        keep = list(filter(lambda s: isinstance(s, str), x))
        if len(keep) == 0:  # pragma: no cover
            return numpy.nan
        keep.sort()
        val = keep[-1]
        if len(val) > 10:  # pragma: no cover
            val = val[:10] + '...'
        return f"{val!r}"
    except (ValueError, TypeError):  # pragma: no cover
        return '?'


def guess_proto_dtype(dtype):
    """
    Guesses the ONNX dtype given a numpy dtype.

    :param dtype: numpy dtype
    :return: proto type
    """
    if dtype == numpy.float32:
        return TensorProto.FLOAT  # pylint: disable=E1101
    if dtype == numpy.float64:
        return TensorProto.DOUBLE  # pylint: disable=E1101
    if dtype == numpy.int64:
        return TensorProto.INT64  # pylint: disable=E1101
    if dtype == numpy.int32:
        return TensorProto.INT32  # pylint: disable=E1101
    if dtype == numpy.int16:
        return TensorProto.INT16  # pylint: disable=E1101
    if dtype == numpy.int8:
        return TensorProto.INT8  # pylint: disable=E1101
    if dtype == numpy.uint64:
        return TensorProto.UINT64  # pylint: disable=E1101
    if dtype == numpy.uint32:
        return TensorProto.UINT32  # pylint: disable=E1101
    if dtype == numpy.uint16:
        return TensorProto.UINT16  # pylint: disable=E1101
    if dtype == numpy.uint8:
        return TensorProto.UINT8  # pylint: disable=E1101
    if dtype == numpy.float16:
        return TensorProto.FLOAT16  # pylint: disable=E1101
    if dtype in (bool, numpy.bool_):
        return TensorProto.BOOL  # pylint: disable=E1101
    if dtype in (str, numpy.str_):
        return TensorProto.STRING  # pylint: disable=E1101
    raise RuntimeError(
        f"Unable to guess type for dtype={dtype}.")  # pragma: no cover


def guess_proto_dtype_name(onnx_dtype):
    """
    Returns a string equivalent to `onnx_dtype`.

    :param dtype: onnx dtype
    :return: proto type
    """
    if onnx_dtype == TensorProto.FLOAT:  # pylint: disable=E1101
        return "TensorProto.FLOAT"
    if onnx_dtype == TensorProto.DOUBLE:  # pylint: disable=E1101
        return "TensorProto.DOUBLE"
    if onnx_dtype == TensorProto.INT64:  # pylint: disable=E1101
        return "TensorProto.INT64"
    if onnx_dtype == TensorProto.INT32:  # pylint: disable=E1101
        return "TensorProto.INT32"
    if onnx_dtype == TensorProto.INT16:  # pylint: disable=E1101
        return "TensorProto.INT16"
    if onnx_dtype == TensorProto.UINT8:  # pylint: disable=E1101
        return "TensorProto.UINT8"
    if onnx_dtype == TensorProto.FLOAT16:  # pylint: disable=E1101
        return "TensorProto.FLOAT16"
    if onnx_dtype == TensorProto.BFLOAT16:  # pylint: disable=E1101
        return "TensorProto.BFLOAT16"
    if onnx_dtype == TensorProto.BOOL:  # pylint: disable=E1101
        return "TensorProto.BOOL"
    if onnx_dtype == TensorProto.STRING:  # pylint: disable=E1101
        return "TensorProto.STRING"
    raise RuntimeError(  # pragma: no cover
        f"Unable to guess type for dtype={onnx_dtype}.")


def guess_dtype(proto_type):
    """
    Converts a proto type into a :epkg:`numpy` type.

    :param proto_type: example ``onnx.TensorProto.FLOAT``
    :return: :epkg:`numpy` dtype
    """
    if proto_type == TensorProto.FLOAT:  # pylint: disable=E1101
        return numpy.float32
    if proto_type == TensorProto.BOOL:  # pylint: disable=E1101
        return numpy.bool_
    if proto_type == TensorProto.DOUBLE:  # pylint: disable=E1101
        return numpy.float64
    if proto_type == TensorProto.STRING:  # pylint: disable=E1101
        return numpy.str_
    if proto_type == TensorProto.INT64:  # pylint: disable=E1101
        return numpy.int64
    if proto_type == TensorProto.INT32:  # pylint: disable=E1101
        return numpy.int32
    if proto_type == TensorProto.INT8:  # pylint: disable=E1101
        return numpy.int8
    if proto_type == TensorProto.INT16:  # pylint: disable=E1101
        return numpy.int16
    if proto_type == TensorProto.UINT64:  # pylint: disable=E1101
        return numpy.uint64
    if proto_type == TensorProto.UINT32:  # pylint: disable=E1101
        return numpy.uint32
    if proto_type == TensorProto.UINT8:  # pylint: disable=E1101
        return numpy.uint8
    if proto_type == TensorProto.UINT16:  # pylint: disable=E1101
        return numpy.uint16
    if proto_type == TensorProto.FLOAT16:  # pylint: disable=E1101
        return numpy.float16
    raise ValueError(
        f"Unable to convert proto_type {proto_type} to numpy type.")


def to_skl2onnx_type(name, elem_type, shape):
    """
    Converts *name*, *elem_type*, *shape* into a
    :epkg:`sklearn-onnx` type.

    :param name: string
    :param elem_type: tensor of elements of this type
    :param shape: expected shape
    :return: data type
    """
    from skl2onnx.common.data_types import _guess_numpy_type  # delayed
    elem = guess_numpy_type_from_string(elem_type)
    shape = list(None if d == 0 else d for d in shape)
    return (name, _guess_numpy_type(elem, shape))


def from_pb(obj):
    """
    Extracts tensor description from a protobuf.

    :param obj: initializer, tensor
    :return: (name, type, shape)
    """
    def get_dim(d):
        r = d.dim_value
        if "dim_param" in str(d):
            return None
        if r == 0:
            # dim_value is 0 when it is 0 or undefined
            return 0 if "0" in str(d) else None
        return r

    def get_shape(tt):
        return [get_dim(tt.shape.dim[i])
                for i in range(len(tt.shape.dim))]

    if hasattr(obj, 'extend'):
        return [from_pb(o) for o in obj]

    name = obj.name
    if obj.type.tensor_type:
        tt = obj.type.tensor_type
        elem = tt.elem_type
        shape = get_shape(tt)
        if elem not in TENSOR_TYPE_TO_NP_TYPE:
            raise NotImplementedError(
                f"Unsupported type '{type(obj.type.tensor_type)}' (elem_type={elem}).")
        ty = TENSOR_TYPE_TO_NP_TYPE[elem].type
    else:
        raise NotImplementedError(  # pragma: no cover
            f"Unsupported type '{type(obj)}' as a string ({obj}).")

    return (name, ty, shape)


def numpy_type_prototype(dtype):
    """
    Converts a numpy dtyp into a TensorProto dtype.

    :param dtype: dtype
    :return: proto dtype
    """
    if dtype in NP_TYPE_TO_TENSOR_TYPE:
        return NP_TYPE_TO_TENSOR_TYPE[dtype]
    dt = numpy.dtype(dtype)
    if dt in NP_TYPE_TO_TENSOR_TYPE:
        return NP_TYPE_TO_TENSOR_TYPE[dt]
    raise ValueError(  # pragma: no cover
        f"Unable to convert dtype {dtype!r} into ProtoType.")


def make_value_info(name, dtype, shape):
    """
    Converts a variable defined by its name, type and shape
    into `onnx.ValueInfoProto`.

    :param name: name
    :param dtype: numpy element type
    :param shape: shape
    :return: instance of `onnx.ValueInfoProto`
    """
    value_info = ValueInfoProto()
    value_info.name = name
    tensor_type_proto = make_tensor_type_proto(
        numpy_type_prototype(dtype), shape)
    value_info.type.CopyFrom(tensor_type_proto)  # pylint: disable=E1101
    return value_info


def copy_value_info(info, name=None):
    """
    Makes a copy of `onnx.ValueInfoProto`.

    :param name: if defined, changed the name
    :return: instance of `onnx.ValueInfoProto`
    """
    value_info = ValueInfoProto()
    value_info.name = name or info.name
    value_info.type.CopyFrom(info.type)  # pylint: disable=E1101
    return value_info


_get_onnx_function_cache = None


def _get_onnx_function():
    """
    Returns the list of functions defined in ONNX package.
    """
    global _get_onnx_function_cache  # pylint: disable=W0603
    if _get_onnx_function_cache is None:
        _get_onnx_function_cache = {}
        fcts = get_function_ops()
        for fct in fcts:
            key = fct.domain, fct.name
            if key in _get_onnx_function_cache:
                raise RuntimeError(  # pragma: no cover
                    f"Function {key!r} is already registered.")
            _get_onnx_function_cache[key] = fct
    return _get_onnx_function_cache


def get_onnx_schema(opname, domain='', opset=None, load_function=False):
    """
    Returns the operator schema for a specific operator.

    :param domain: operator domain
    :param opname: operator name
    :param opset: opset or version, None for the latest
    :param load_function: loads the function, if True, the function
        looks into the list of function if one of them has the same name,
        opset must be None in that case
    :return: :epkg:`OpSchema`
    """
    if load_function:
        if opset is not None:
            raise ValueError(
                "opset must be None if load_function is True for "
                "operator (%r,%r)." % (domain, opname))
        fcts = _get_onnx_function()
        key = domain, opname
        if key in fcts:
            return fcts[key]
        if opset is None:
            opset = onnx_opset_version()
        return get_schema(opname, opset, domain)
    if opset is None:
        opset = onnx_opset_version()
    return get_schema(opname, opset, domain)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class ZipMapDictionary(dict):
    """
    Custom dictionary class much faster for this runtime,
    it implements a subset of the same methods.
    """
    __slots__ = ['_rev_keys', '_values', '_mat']

    @staticmethod
    def build_rev_keys(keys):
        res = {}
        for i, k in enumerate(keys):
            res[k] = i
        return res

    def __init__(self, rev_keys, values, mat=None):
        """
        @param      rev_keys        returns by @see me build_rev_keys,
                                    *{keys: column index}*
        @param      values          values
        @param      mat             matrix if values is a row index,
                                    one or two dimensions
        """
        if mat is not None:
            if not isinstance(mat, numpy.ndarray):
                raise TypeError(  # pragma: no cover
                    f'matrix is expected, got {type(mat)}.')
            if len(mat.shape) not in (2, 3):
                raise ValueError(  # pragma: no cover
                    f"matrix must have two or three dimensions but got {mat.shape}.")
        dict.__init__(self)
        self._rev_keys = rev_keys
        self._values = values
        self._mat = mat

    def __getstate__(self):
        """
        For pickle.
        """
        return dict(_rev_keys=self._rev_keys,
                    _values=self._values,
                    _mat=self._mat)

    def __setstate__(self, state):
        """
        For pickle.
        """
        if isinstance(state, tuple):
            state = state[1]
        self._rev_keys = state['_rev_keys']
        self._values = state['_values']
        self._mat = state['_mat']

    def __getitem__(self, key):
        """
        Returns the item mapped to keys.
        """
        if self._mat is None:
            return self._values[self._rev_keys[key]]
        return self._mat[self._values, self._rev_keys[key]]

    def __setitem__(self, pos, value):
        "unused but used by pickle"
        pass

    def __len__(self):
        """
        Returns the number of items.
        """
        return len(self._values) if self._mat is None else self._mat.shape[1]

    def __iter__(self):
        for k in self._rev_keys:
            yield k

    def __contains__(self, key):
        return key in self._rev_keys

    def items(self):
        if self._mat is None:
            for k, v in self._rev_keys.items():
                yield k, self._values[v]
        else:
            for k, v in self._rev_keys.items():
                yield k, self._mat[self._values, v]

    def keys(self):
        for k in self._rev_keys.keys():
            yield k

    def values(self):
        if self._mat is None:
            for v in self._values:
                yield v
        else:
            for v in self._mat[self._values]:
                yield v

    def asdict(self):
        res = {}
        for k, v in self.items():
            res[k] = v
        return res

    def __str__(self):
        return f"ZipMap({str(self.asdict())!r})"


class ArrayZipMapDictionary(list):
    """
    Mocks an array without changing the data it receives.
    Notebooks :ref:`onnxnodetimerst` illustrates the weaknesses
    and the strengths of this class compare to a list
    of dictionaries.

    .. index:: ZipMap
    """

    def __init__(self, rev_keys, mat):
        """
        @param      rev_keys        dictionary *{keys: column index}*
        @param      mat             matrix if values is a row index,
                                    one or two dimensions
        """
        if mat is not None:
            if not isinstance(mat, numpy.ndarray):
                raise TypeError(  # pragma: no cover
                    f'matrix is expected, got {type(mat)}.')
            if len(mat.shape) not in (2, 3):
                raise ValueError(  # pragma: no cover
                    f"matrix must have two or three dimensions but got {mat.shape}.")
        list.__init__(self)
        self._rev_keys = rev_keys
        self._mat = mat

    @property
    def dtype(self):
        return self._mat.dtype

    def __len__(self):
        return self._mat.shape[0]

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, i):
        return ZipMapDictionary(self._rev_keys, i, self._mat)

    def __setitem__(self, pos, value):
        raise RuntimeError(
            f"Changing an element is not supported (pos=[{pos}]).")

    @property
    def values(self):
        """
        Equivalent to ``DataFrame(self).values``.
        """
        if len(self._mat.shape) == 3:
            return self._mat.reshape((self._mat.shape[1], -1))
        return self._mat

    @property
    def columns(self):
        """
        Equivalent to ``DataFrame(self).columns``.
        """
        res = [(v, k) for k, v in self._rev_keys.items()]
        if len(res) == 0:
            if len(self._mat.shape) == 2:
                res = [(i, 'c%d' % i) for i in range(self._mat.shape[1])]
            elif len(self._mat.shape) == 3:
                # multiclass
                res = [(i, 'c%d' % i)
                       for i in range(self._mat.shape[0] * self._mat.shape[2])]
            else:
                raise RuntimeError(  # pragma: no cover
                    "Unable to guess the right number of columns for "
                    "shapes: {}".format(self._mat.shape))
        else:
            res.sort()
        return [_[1] for _ in res]

    @property
    def is_zip_map(self):
        return True

    def __str__(self):
        return f"ZipMaps[{', '.join(map(str, self))}]"


class ZipMap(OpRun):
    """
    The class does not output a dictionary as
    specified in :epkg:`ONNX` specifications
    but a @see cl ArrayZipMapDictionary which
    is wrapper on the input so that it does not
    get copied.
    """

    atts = {'classlabels_int64s': [], 'classlabels_strings': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=ZipMap.atts,
                       **options)
        if hasattr(self, 'classlabels_int64s') and len(self.classlabels_int64s) > 0:
            self.rev_keys_ = ZipMapDictionary.build_rev_keys(
                self.classlabels_int64s)
        elif hasattr(self, 'classlabels_strings') and len(self.classlabels_strings) > 0:
            self.rev_keys_ = ZipMapDictionary.build_rev_keys(
                self.classlabels_strings)
        else:
            self.rev_keys_ = {}

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        res = ArrayZipMapDictionary(self.rev_keys_, x)
        return (res, )



"""
@file
@brief Functions to help exporting json format into text.
"""
import pprint
import copy
import os
import json
from json.decoder import JSONDecodeError


def fix_missing_imports():
    """
    The execution of a file through function :epkg:`exec`
    does not import new modules. They must be there when
    it is done. This function fills the gap for some of
    them.

    @return         added names
    """
    from sklearn.linear_model import LogisticRegression
    return {'LogisticRegression': LogisticRegression}


def _dict2str(d):
    vals = []
    for k, v in d.items():
        if isinstance(v, dict):
            vals.append(f"{k}{_dict2str(v)}")
        else:
            vals.append(f"{k}{v}")
    return "-".join(vals)


def _coor_to_str(cc):
    ccs = []
    for c in cc:
        if c in ('{}', {}):
            c = "o"
        elif len(c) > 1 and (c[0], c[-1]) == ('{', '}'):
            c = c.replace("<class ", "")
            c = c.replace(">:", ":")
            c = c.replace("'", '"').replace("True", "1").replace("False", "0")
            try:
                d = json.loads(c)
            except JSONDecodeError as e:  # pragma: no cover
                raise RuntimeError(
                    f"Unable to interpret '{c}'.") from e

            if len(d) == 1:
                its = list(d.items())[0]
                if '.' in its[0]:
                    c = _dict2str(its[1])
                else:
                    c = _dict2str(d)
            else:
                c = _dict2str(d)
        c = str(c).strip("'")
        ccs.append(c)
    return 'M-' + "-".join(map(str, ccs)).replace("'", "")


def _figures2dict(metrics, coor, baseline=None):
    """
    Converts the data from list to dictionaries.

    @param      metrics     single array of values
    @param      coor        list of list of coordinates names
    @param      baseline    one coordinates is the baseline
    @return                 dictionary of metrics
    """
    if baseline is None:
        base_j = None
    else:
        quoted_base = f"'{baseline}'"
        base_j = None
        for i, base in enumerate(coor):
            if baseline in base:
                base_j = i, base.index(baseline)
                break
            if quoted_base in base:
                base_j = i, base.index(quoted_base)
                break
        if base_j is None:
            raise ValueError(  # pragma: no cover
                "Unable to find value baseline '{}' or [{}] in {}".format(
                    baseline, quoted_base, pprint.pformat(coor)))
    m_bases = {}
    ind = [0 for c in coor]
    res = {}
    pos = 0
    while ind[0] < len(coor[0]):
        cc = [coor[i][ind[i]] for i in range(len(ind))]
        if baseline is not None:
            if cc[base_j[0]] != base_j[1]:
                cc2 = cc.copy()
                cc2[base_j[0]] = coor[base_j[0]][base_j[1]]
                key = tuple(cc2)
                skey = _coor_to_str(key)
                if key not in m_bases:
                    m_bases[skey] = []
                m_bases[skey].append(_coor_to_str(cc))

        name = _coor_to_str(cc)
        res[name] = metrics[pos]
        pos += 1
        ind[-1] += 1
        last = len(ind) - 1
        while last > 0 and ind[last] >= len(coor[last]):
            ind[last] = 0
            last -= 1
            ind[last] += 1

    for k, v in m_bases.items():
        for ks in v:
            if (k in res and res[k] != 0 and ks in res and
                    res[ks] is not None and res[k] is not None):
                res['R-' + ks[2:]] = float(res[ks]) / res[k]
    return res


def enumerate_export_asv_json(folder, as_df=False, last_one=False,
                              baseline=None, conf=None):
    """
    Looks into :epkg:`asv` results and wraps all of them
    into a :epkg:`dataframe` or flat data.

    @param      folder      location of the results
    @param      last_one    to return only the last one
    @param      baseline    defines a baseline and computes ratios
    @param      conf        configuration file, may be used to
                            add additional data
    @return                 :epkg:`dataframe` or flat data
    """
    meta_class = None
    if conf is not None:
        if not os.path.exists(conf):
            raise FileNotFoundError(  # pragma: no cover
                f"Unable to find '{conf}'.")
        with open(conf, "r", encoding='utf-8') as f:
            meta = json.load(f)
        bdir = os.path.join(os.path.dirname(conf), meta['benchmark_dir'])
        if os.path.exists(bdir):
            meta_class = _retrieve_class_parameters(bdir)

    bench = os.path.join(folder, 'benchmarks.json')
    if not os.path.exists(bench):
        raise FileNotFoundError(  # pragma: no cover
            f"Unable to find '{bench}'.")
    with open(bench, 'r', encoding='utf-8') as f:
        content = json.load(f)

    # content contains the list of tests
    content = {k: v for k, v in content.items() if isinstance(v, dict)}

    # looking into metadata
    machines = os.listdir(folder)
    for machine in machines:
        if 'benchmarks.json' in machine:
            continue
        filemine = os.path.join(folder, machine, 'machine.json')
        with open(filemine, 'r', encoding='utf-8') as f:
            meta = json.load(f)

        # looking into all tests or the last one
        subs = os.listdir(os.path.join(folder, machine))
        subs = [m for m in subs if m != 'machine.json']
        if last_one:
            dates = [(os.stat(os.path.join(folder, machine, m)).st_ctime, m)
                     for m in subs if ('-env' in m or 'virtualenv-' in m) and '.json' in m]
            dates.sort()
            subs = [dates[-1][-1]]

        # look into tests
        for sub in subs:
            data = os.path.join(folder, machine, sub)
            with open(data, 'r', encoding='utf-8') as f:
                test_content = json.load(f)
            meta_res = copy.deepcopy(meta)
            for k, v in test_content.items():
                if k != 'results':
                    meta_res[k] = v
                    continue
                results = test_content['results']
                for kk, vv in results.items():
                    if vv is None:
                        raise RuntimeError(  # pragma: no cover
                            'Unexpected empty value for vv')
                    try:
                        metrics, coord, hash = vv[:3]
                    except ValueError as e:  # pragma: no cover
                        raise ValueError(
                            f"Test '{kk}', unable to interpret: {vv}.") from e

                    obs = {}
                    for mk, mv in meta_res.items():
                        if mk in {'result_columns'}:
                            continue
                        if isinstance(mv, dict):
                            for mk2, mv2 in mv.items():
                                obs[f'{mk}_{mk2}'] = mv2
                        else:
                            obs[mk] = mv
                    spl = kk.split('.')
                    obs['test_hash'] = hash
                    obs['test_fullname'] = kk
                    if len(spl) >= 4:
                        obs['test_model_set'] = spl[0]
                        obs['test_model_kind'] = spl[1]
                        obs['test_model'] = ".".join(spl[2:-1])
                        obs['test_name'] = spl[-1]
                    elif len(spl) >= 3:
                        obs['test_model_set'] = spl[0]
                        obs['test_model'] = ".".join(spl[1:-1])
                        obs['test_name'] = spl[-1]
                    else:
                        obs['test_model'] = ".".join(spl[:-1])
                        obs['test_name'] = spl[-1]
                    if metrics is not None:
                        obs.update(
                            _figures2dict(metrics, coord, baseline=baseline))
                    if meta_class is not None:
                        _update_test_metadata(obs, meta_class)
                    yield obs


def export_asv_json(folder, as_df=False, last_one=False, baseline=None,
                    conf=None):
    """
    Looks into :epkg:`asv` results and wraps all of them
    into a :epkg:`dataframe` or flat data.

    @param      folder      location of the results
    @param      as_df       returns a dataframe or
                            a list of dictionaries
    @param      last_one    to return only the last one
    @param      baseline    computes ratio against the baseline
    @param      conf        configuration file, may be used to
                            add additional data
    @return                 :epkg:`dataframe` or flat data
    """
    rows = list(enumerate_export_asv_json(
        folder, last_one=last_one, baseline=baseline, conf=conf))
    if as_df:
        import pandas
        return pandas.DataFrame(rows)
    return rows


def _retrieve_class_parameters(bdir):
    """
    Imports files in bdir, compile files and extra metadata from them.
    """
    found = {}
    for path, _, files in os.walk(os.path.abspath(bdir)):
        fulls = [os.path.join(path, f) for f in files]
        for full in fulls:
            if (os.path.splitext(full)[-1] == '.py' and
                    os.path.split(full)[-1] != '__init__.py'):
                cls = list(_enumerate_classes(full))
                for cl in cls:
                    name = cl.__name__
                    found[name] = cl
    return found


def _update_test_metadata(row, class_meta):
    name = row.get('test_model', None)
    if name is None:
        return
    sub = name.split('.')[-1]
    if sub in class_meta:
        for k, v in class_meta[sub].__dict__.items():
            if k.startswith('par_'):
                row[k] = v


def _enumerate_classes(filename):
    """
    Extracts the classes of a file.
    """
    with open(filename, "r", encoding="utf-8") as f:
        content = f.read()
    gl = fix_missing_imports()
    loc = {}
    cp = compile(content, filename, mode='exec')

    try:
        exec(cp, gl, loc)  # pylint: disable=W0122
    except NameError as e:  # pragma: no cover
        raise NameError(
            "An import is probably missing from function 'fix_missing_imports'"
            ".") from e

    for k, v in loc.items():
        if k[0] < 'A' or k[0] > 'Z':
            continue
        if not hasattr(v, 'setup_cache'):
            continue
        yield v



"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def softmaxcrossentropy(x, target, weight=None, reduction='mean',
                        ignore_index=None, get_log_prob=None):
    """
    Modified version of `softmaxcrossentropy.py
    <https://github.com/onnx/onnx/blob/main/onnx/backend/
    test/case/node/softmaxcrossentropy.py>`_ to handle other type
    than float32.
    """
    input_shape = x.shape
    if len(input_shape) == 1:
        raise RuntimeError(f"Unsupported shape {input_shape!r}.")

    target_shape = target.shape
    N = input_shape[0]
    C = input_shape[1]

    # compute log_softmax
    max_x = numpy.max(x, axis=1, keepdims=True)
    exp_x = numpy.exp(x - max_x)
    p = exp_x / numpy.sum(exp_x, axis=1, keepdims=True)
    inp = numpy.log(p)
    log_prob = None
    if get_log_prob is True:
        log_prob = numpy.copy(inp)

    # initialize the positional weights when required
    gather_weight = None
    if weight is not None:
        gather_weight = numpy.take(
            weight, numpy.array(target, dtype=numpy.int32), mode='clip')
        if ignore_index is not None:
            gather_weight = numpy.where(
                target == ignore_index, 0, gather_weight).astype(dtype=x.dtype)
    elif ignore_index is not None:
        gather_weight = numpy.where(
            target == ignore_index, 0, 1).astype(dtype=x.dtype)

    # if input is 4-d and above, make it 3-d
    if len(input_shape) != 3:
        inp = inp.reshape((N, C, -1))
        target = target.reshape((N, -1))

    # Get a dimension from the reshaped input.
    # If the original input shape is [N, C, H, W],
    # the D here should be H * W because we reshape
    # [N, C, H, W] to [N, C, H * W].
    D = inp.shape[2]
    neg_gather_element_input = numpy.zeros((N, D), dtype=x.dtype)
    for i in range(N):
        for d in range(D):
            if target[i, d] != ignore_index:
                neg_gather_element_input[i, d] = -inp[i, target[i, d], d]

    loss = neg_gather_element_input

    # if the input was 4-d or above reshape to the right shape
    if len(input_shape) != 3:
        loss = loss.reshape(target_shape)

    # apply the weights when required
    if gather_weight is not None:
        loss = gather_weight * loss
        if reduction == b'mean':
            loss = loss.sum() / gather_weight.sum()
            if get_log_prob is True:
                return loss, log_prob
            return (loss, )

    if reduction == b'mean':
        loss = numpy.mean(loss)
    elif reduction == b'sum':
        loss = numpy.sum(loss)

    if get_log_prob is True:
        return loss, log_prob
    return (loss, )


class SoftmaxCrossEntropyLoss(OpRun):
    """
    Python runtime for function *SoftmaxCrossEntropyLoss*.
    """

    atts = {'reduction': b'mean', 'ignore_index': -1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=SoftmaxCrossEntropyLoss.atts,
                       **options)

    def _run(self, x, target, weight=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        n_outputs = len(self.onnx_node.output)
        return softmaxcrossentropy(
            x, target, weight=weight, reduction=self.reduction,  # pylint: disable=E1101
            ignore_index=self.ignore_index,  # pylint: disable=E1101
            get_log_prob=n_outputs == 2)



# -*- coding: utf-8 -*-
"""
@file
@brief Tiny helpers for scikit-learn exporters.
"""


def check_type(model, model_type):
    """
    Raises an exception if the model is not of the expected type.

    @param      model       *scikit-learn* model
    @param      model_type  expected type
    """
    if not isinstance(model, model_type):
        raise TypeError(  # pragma: no cover
            f"Model type {type(model)} is not of type {model_type}.")



"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _compute_negative_log_likelihood_loss(x, target, weight=None,
                                          reduction=b'mean', ignore_index=None):
    """
    Modified version of `softmaxcrossentropy.py
    <https://github.com/onnx/onnx/blob/main/onnx/backend/
    test/case/node/negativeloglikelihoodloss.py>`_ to handle other type
    than float32.
    """
    input_shape = x.shape
    if len(input_shape) == 1:
        raise RuntimeError(f"Unsupported shape {input_shape!r}.")

    target_shape = target.shape
    N = input_shape[0]
    C = input_shape[1]

    # initialize the positional weights when required
    gather_weight = None
    if weight is not None:
        # setting mode='clip' to deal with ignore_index > C or < 0 cases.
        # when the target value is > C or < 0, it doesn't matter which value we are
        # taking in gather_weight, since it will be set to 0 in the following if-block
        # use numpy.int32 to make it compatible with x86 machines
        gather_weight = numpy.take(weight, numpy.array(
            target, dtype=numpy.int32), mode='clip')
        # set `ignore_index`'s loss weight to 0.
        # The loss tensor will be multiplied by this weight tensor,
        # so `ingore_index`'s loss value will be eliminated.
        if ignore_index is not None:
            gather_weight = numpy.where(
                target == ignore_index, 0, gather_weight).astype(dtype=x.dtype)
    elif ignore_index != -1:
        gather_weight = numpy.where(
            target == ignore_index, 0, 1).astype(dtype=x.dtype)

    # if input is 4-d and above, make it 3-d
    if len(input_shape) != 3:
        x = x.reshape((N, C, -1))
        target = target.reshape((N, -1))

    # Get a dimension from the reshaped input.
    # If the original input shape is [N, C, H, W],
    # the D here should be H * W because we reshape
    # [N, C, H, W] to [N, C, H * W].
    D = x.shape[2]
    neg_gather_element_input = numpy.zeros((N, D), dtype=x.dtype)
    for i in range(N):
        for d in range(D):
            if target[i][d] != ignore_index:
                neg_gather_element_input[i][d] = -x[i][target[i][d]][d]

    loss = neg_gather_element_input

    # if the input was 4-d or above reshape to the right shape
    if len(input_shape) != 3:
        loss = loss.reshape(target_shape)

    # apply the weights when required
    if gather_weight is not None:
        loss = gather_weight * loss
        if reduction == b'mean':
            loss = loss.sum() / gather_weight.sum()
            return (loss, )

    if reduction == b'mean':
        loss = numpy.mean(loss)
    elif reduction == b'sum':
        loss = numpy.sum(loss)
    return (loss, )


class NegativeLogLikelihoodLoss(OpRun):
    """
    Python runtime for function *NegativeLogLikelihoodLoss*.
    """

    atts = {'reduction': b'mean', 'ignore_index': -1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=NegativeLogLikelihoodLoss.atts,
                       **options)

    def _run(self, x, target, weight=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return _compute_negative_log_likelihood_loss(
            x, target, weight=weight, reduction=self.reduction,  # pylint: disable=E1101
            ignore_index=self.ignore_index)  # pylint: disable=E1101



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import itertools
import numpy
from ._op import OpRun


def _get_pad_shape(auto_pad, input_spatial_shape, kernel_spatial_shape,
                   strides_spatial, output_spatial_shape):
    pad_shape = [0] * len(input_spatial_shape)
    if auto_pad in ('SAME_UPPER', 'SAME_LOWER'):
        for i in range(len(input_spatial_shape)):  # pylint: disable=C0200
            pad_shape[i] = (
                (output_spatial_shape[i] - 1) * strides_spatial[i] +
                kernel_spatial_shape[i] - input_spatial_shape[i])
    elif auto_pad == 'VALID':
        pass
    if len(pad_shape) == 0:
        raise RuntimeError(  # pragma: no cover
            "Unable to compute pad shape, auto_pad=%r, "
            "input_spatial_shape=%r, kernel_spatial_shape=%r, "
            "strides_spatial=%r." % (
                auto_pad, input_spatial_shape, kernel_spatial_shape,
                strides_spatial))
    return pad_shape


def _get_output_shape_no_ceil(auto_pad, input_spatial_shape, kernel_spatial_shape,
                              strides_spatial):
    out_shape = [0] * len(input_spatial_shape)
    if auto_pad in ('SAME_UPPER', 'SAME_LOWER'):
        for i in range(len(input_spatial_shape)):  # pylint: disable=C0200
            out_shape[i] = int(
                numpy.ceil(
                    float(input_spatial_shape[i]) /
                    float(strides_spatial[i])))
    elif auto_pad == 'VALID':
        for i in range(len(input_spatial_shape)):  # pylint: disable=C0200
            out_shape[i] = int(
                numpy.ceil(
                    float(input_spatial_shape[i] -
                          (kernel_spatial_shape[i] - 1)) /
                    float(strides_spatial[i])))
    return out_shape


def _get_output_shape(auto_pad, input_spatial_shape, kernel_spatial_shape,
                      strides_spatial, pad_shape=None, ceil_mode=0):
    if not ceil_mode:
        out_shape = _get_output_shape_no_ceil(
            auto_pad, input_spatial_shape, kernel_spatial_shape,
            strides_spatial)
    else:
        round_fct = numpy.ceil if ceil_mode else numpy.floor
        out_shape = [0] * len(input_spatial_shape)
        if auto_pad in ('SAME_UPPER', 'SAME_LOWER'):
            for i in range(len(input_spatial_shape)):  # pylint: disable=C0200
                out_shape[i] = int(
                    round_fct(float(input_spatial_shape[i]) / float(strides_spatial[i])))
        elif auto_pad == 'VALID':
            if pad_shape is None:
                raise ValueError(  # pragma: no cover
                    "pad_shape cannot be None if auto_pad is "
                    "'VALID' and ceil_mode is 1.")
            for i in range(len(input_spatial_shape)):  # pylint: disable=C0200
                out_shape[i] = int(
                    round_fct(
                        float(input_spatial_shape[i] + pad_shape[i] - kernel_spatial_shape[i]) /
                        float(strides_spatial[i]) + 1))
    if len(out_shape) == 0:
        raise RuntimeError(  # pragma: no cover
            "Unable to compute output shape, auto_pad=%r, "
            "input_spatial_shape=%r, kernel_spatial_shape=%r, "
            "strides_spatial=%r, ceil_mode=%r." % (
                auto_pad, input_spatial_shape, kernel_spatial_shape,
                strides_spatial, ceil_mode))
    if min(out_shape) <= 0:
        raise RuntimeError(  # pragma: no cover
            "output shape cannot be null or negative, out_shape=%r, "
            "auto_pad=%r, input_spatial_shape=%r, "
            "kernel_spatial_shape=%r, strides_spatial=%r, ceil_mode=%r." % (
                out_shape, auto_pad, input_spatial_shape,
                kernel_spatial_shape, strides_spatial, ceil_mode))
    return out_shape


def _pool(padded, x_shape, kernel_shape, strides_shape,
          out_shape, pad_shape, pooling_type, count_include_pad=0, ceil_mode=0):
    if pooling_type == 'AVG':
        fpool = numpy.average
    elif pooling_type == 'MAX':
        fpool = numpy.max
    else:
        raise NotImplementedError(  # pragma: no cover
            f'Pooling type {pooling_type} does not support. Should be AVG, MAX.')
    spatial_size = len(x_shape) - 2
    y = numpy.zeros([x_shape[0], x_shape[1]] + list(out_shape))
    round_fct = numpy.ceil if ceil_mode else numpy.floor

    def loop_range():
        return [range(int(round_fct(
                float(x_shape[i + 2] + pad_shape[i] - kernel_shape[i]) /
                float(strides_shape[i]) + 1))) for i in range(spatial_size)]

    for shape in itertools.product(range(x_shape[0]), range(x_shape[1]), *loop_range()):
        window = padded[shape[0], shape[1]]
        listi = [range(strides_shape[i] * shape[i + 2],
                       strides_shape[i] * shape[i + 2] + kernel_shape[i])
                 for i in range(spatial_size)]
        listi2 = list(itertools.product(*listi))
        values = []
        for i in listi2:
            try:
                values.append(window[i])
            except IndexError:
                continue
        window_vals = numpy.array(values)

        if count_include_pad == 1 and pooling_type == 'AVG':
            y[shape] = fpool(window_vals)
        else:
            y[shape] = fpool(
                window_vals[numpy.where(~numpy.isnan(window_vals))])
    return y.astype(numpy.float32)


class AveragePool(OpRun):

    atts = {'auto_pad': b'NOTSET',
            'ceil_mode': 0,
            'count_include_pad': 0,
            'kernel_shape': [],
            'pads': [],
            'strides': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=AveragePool.atts,
                       **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(self.strides) == 0:
            strides = [1] * (len(x.shape) - 2)
        else:
            strides = self.strides
        kernel_shape = list(self.kernel_shape)
        auto_pad = (
            'VALID' if self.auto_pad == b'NOTSET'
            else self.auto_pad.decode('ascii'))

        if len(self.pads) == 0:
            pad_shape = [0] * (len(x.shape) - 2)
            x_shape = x.shape[2:]
            padded = x
        elif len(self.pads) == 4:
            pad_top, pad_bottom, pad_left, pad_right = self.pads
            pad_shape = [pad_top + pad_bottom, pad_left + pad_right]
            x_shape = numpy.array(x.shape[2:]) + numpy.array(pad_shape)
            const = numpy.nan if self.count_include_pad == 0 else 0
            padded = numpy.pad(
                x, ((0, 0), (0, 0),
                    (pad_top, pad_bottom), (pad_left, pad_right)),
                mode='constant', constant_values=const)
        else:
            pad_shape = self.pads
            x_shape = x.shape[2:]
            padded = x

        if auto_pad in ('SAME_LOWER', 'SAME_UPPER'):
            const = numpy.nan if self.count_include_pad == 0 else 0
            out_shape = _get_output_shape(
                auto_pad, x_shape, kernel_shape, strides, pad_shape, self.ceil_mode)
            pad_shape = _get_pad_shape(
                auto_pad, x_shape, kernel_shape, strides, out_shape)
            if auto_pad == 'SAME_LOWER':
                pad_bottom = pad_shape[0] // 2
                pad_top = pad_shape[0] - pad_bottom
                pad_right = pad_shape[1] // 2
                pad_left = pad_shape[1] - pad_right
            else:
                pad_top = pad_shape[0] // 2
                pad_bottom = pad_shape[0] - pad_top
                pad_left = pad_shape[1] // 2
                pad_right = pad_shape[1] - pad_left
            padded = numpy.pad(
                padded, ((0, 0), (0, 0), (pad_top, pad_bottom),
                         (pad_left, pad_right)),
                mode='constant', constant_values=const)
        else:
            out_shape = _get_output_shape(
                auto_pad, x_shape, kernel_shape, strides, pad_shape, self.ceil_mode)

        pooling_type = 'AVG'
        res = _pool(padded, x.shape, kernel_shape, strides,
                    out_shape, pad_shape, pooling_type,
                    count_include_pad=self.count_include_pad,
                    ceil_mode=self.ceil_mode)
        return (res, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _scatter_nd_impl(data, indices, updates, reduction=b'none'):
    output = numpy.copy(data)
    for i in numpy.ndindex(indices.shape[:-1]):
        if reduction == 'add':
            output[indices[i]] += updates[i]
        elif reduction == 'mul':
            output[indices[i]] *= updates[i]
        else:
            output[indices[i]] = updates[i]
    return output


class ScatterND(OpRun):

    atts = {'reduction': b'none'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=ScatterND.atts,
                       **options)

    def _run(self, data, indices, updates, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        y = _scatter_nd_impl(data, indices, updates, reduction=self.reduction)
        return (y, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunBinaryNumpy


class BitShift(OpRunBinaryNumpy):

    atts = {'direction': b''}

    def __init__(self, onnx_node, desc=None, **options):
        "constructor"
        OpRunBinaryNumpy.__init__(self, numpy.add, onnx_node,
                                  expected_attributes=BitShift.atts,
                                  desc=desc, **options)
        if self.direction not in (b'LEFT', b'RIGHT'):
            raise ValueError(  # pragma: no cover
                f"Unexpected value for direction ({self.direction!r}).")
        if self.direction == b'LEFT':
            self.numpy_fct = numpy.left_shift
        else:
            self.numpy_fct = numpy.right_shift



"""
@file
@brief Main functions decomposing einsum computation into
more simple functions.
"""
import numpy
from .einsum_impl_classes import EinsumSubOp, GraphEinsumSubOp


def analyse_einsum_equation(equation):
    """
    Analyses an einsum equation.

    :param equation: :epkg:`numpy:einsum` equation
    :return: three results, list of letters,
        a matrix (see below), lengths of each components,
        duplicates

    The returned a matrix is defined as follows:

    .. math::

        m_{ij}=\\left\\{\\begin{array}{ll}-1 &
        \\text{if letter j is involved in input i} \\\\
        p & \\text{p is position of letter j in equation i}
        \\end{array}\\right.
    """
    spl = equation.strip(' ,').split("->")
    if len(spl) != 2 or len(spl[1]) == 0 or len(spl[0]) == 0:
        raise NotImplementedError(
            "The function only implements the case when there are "
            "two sides in the equation: %r." % equation)
    inputs = list(map(lambda s: s.strip(), spl[0].split(',')))
    output = spl[1]
    all_letters = set(inputs[0])

    # Set of letters
    for inp in inputs[1:]:
        all_letters |= set(inp)
    letters = list(sorted(all_letters))
    for c in letters:
        if not(('a' <= c <= 'z') or ('A' <= c <= 'Z')):
            raise ValueError(
                "Equation %r must only contain lower or upper letters "
                "but %r is not." % (equation, c))

    rev = {c: i for i, c in enumerate(letters)}
    for c in output:
        if c not in letters:
            raise ValueError(
                "Output contains one unexpected letter %r in "
                "equation %r." % (c, equation))
    mat = numpy.full((len(inputs) + 1, len(letters)), -1, dtype=numpy.int8)
    for i, inp in enumerate(inputs):
        for k, c in enumerate(inp):
            mat[i, rev[c]] = k
    for k, c in enumerate(output):
        mat[len(inputs), rev[c]] = k
    lengths = [len(inp) for inp in inputs]
    lengths.append(len(output))

    # Look for duplicates
    duplicates = []
    for inp in inputs + [output]:
        if len(inp) == len(set(inp)):
            duplicates.append(None)
            continue
        # There is some duplicates.
        counts = {}
        for i, c in enumerate(inp):
            if c in counts:
                counts[c].append(i)
            else:
                counts[c] = [i]
        duplicates.append(counts)

    return "".join(letters), mat, lengths, duplicates


def decompose_einsum_equation(equation, *shapes, strategy="simple",
                              clean=False, verbose=False):
    """
    Decomposes an equation used in :epkg:`numpy:einsum` knowing
    the input shapes. It returns a sequence of operations
    to do to compute the results.

    :param equation: a string
    :param shapes: sequence of input shapes
    :param strategy: there are different way to decompose the equation,
        this parameters defines the way to do it (see below)
    :param clean: clean the unnecessary node in the graph
    :param verbose: verbosity
    :return: instance of @see cl GraphEinsumSubOp

    About *strategy*:

    * `'simple'`: align all dimensions in the alphabetical order,
      some generic matrix multiplication remains implemented with
      :epkg:`numpy:einsum` but only with two matrices aligned on
      the same dimension (see @see fn numpy_extended_dot)
    * `'numpy'`: same as `simple` but the decomposition does not use
      :epkg:`numpy:einsum` anymore but only multiplication or
      matrix multiplication merged into a single operator called
      *batch_dot* (see @see fn numpy_extended_dot_matrix)

    Available operations: *expand_dims*, *transpose*, *matmul*, *reduce_sum*,
    *id*, *squeeze*, *diagonal*. It analyses an equation and produces a graph
    where node are instance of class @see cl EinsumSubOp.

    .. runpython::
        :showcode:

        from mlprodict.testing.einsum import decompose_einsum_equation
        seq = decompose_einsum_equation("bac,cd,def->ebc")
        for op in seq:
            print(op)

    It can be better displayed as the following.

    .. gdot::
        :script: DOT-SECTION
        :process:

        from mlprodict.testing.einsum import decompose_einsum_equation
        seq = decompose_einsum_equation(
            "bac,cd,def->ebc", (2, 2, 2), (2, 2), (2, 2, 2))
        print("DOT-SECTION", seq.to_dot())

    See notebook :ref:`einsumdecompositionrst`.
    """
    if len(shapes) > 0:
        for sh in shapes:
            if not isinstance(sh, tuple):
                raise TypeError(
                    f"All shapes must be tuples for {sh!r} is not.")
    if strategy in ("simple", "numpy"):
        op_matmul = {'simple': 'matmul',
                     'numpy': 'batch_dot'}
        graph = _decompose_einsum_equation_simple(
            equation, *shapes, verbose=verbose, op_matmul=op_matmul[strategy])
    else:
        raise ValueError(f"Unknown strategy {strategy!r}.")

    # Last step: clean unused nodes.
    if clean:
        last_node = graph.last_added_op
        graph.append(EinsumSubOp(last_node.full_dim, 'id', last_node))
        graph.mark_last_node()
        graph.simplify_mm_nodes(verbose=verbose)
        graph.remove_duplicate_transpose(verbose=verbose)
        graph.clean_unused_nodes(verbose=verbose)
    else:
        graph.mark_last_node()
    return graph


def apply_einsum_sequence(seq, *inputs, verbose=False, **kwargs):
    """
    Applies a sequence of operations on a list of inputs.
    The sequence of operations is produced by function
    @see fn decompose_einsum_equation.

    :param seq: sequence of operations
    :param inputs: inputs
    :param kwargs: additional parameters,
        see :meth:`apply_sequence
        <mlprodict.testing.einsum.einsum_impl_classes.
        GraphEinsumSubOp.apply_sequence>`.
    :return: output

    .. runpython::
        :showcode:

        import numpy
        from mlprodict.testing.einsum import (
            decompose_einsum_equation, apply_einsum_sequence)

        m1 = numpy.arange(2 * 2 * 2).reshape((2, 2, 2)) + 10
        m2 = numpy.arange(4).reshape((2, 2)) + 100
        m3 = numpy.arange(8).reshape((2, 2, 2)) + 1000

        seq = decompose_einsum_equation("bac,cd,def->ebc")
        res = apply_einsum_sequence(seq, m1, m2, m3)
        print(res)

    See notebook :ref:`einsumdecompositionrst`.
    """
    return seq.apply_sequence(*inputs, verbose=verbose, **kwargs)


def is_transpose_identity(perm):
    """
    Tells if the permutation *perm* does nothing (itentity).

    :param perm: permutation
    :return: boolean
    """
    return list(perm) == list(range(len(perm)))


def _basic_verification(lengths, shapes, equation):
    if len(lengths) - 1 != len(shapes):
        raise ValueError(
            "Equation %r has %d inputs but %d shapes are given."
            "" % (equation, len(lengths), len(shapes)))
    for i, (le, sh) in enumerate(zip(lengths, shapes)):
        if le != len(sh):
            raise ValueError(
                "Inputs %d has %d dimensions but shapes %r has %d "
                " in equation %r." % (i, le, sh, len(sh), equation))


def _apply_transpose_reshape(op, row):
    """
    Put all dimensions in the same order.

    :param op: integer (for one input) or an operator
    :param row: letter involved in this input (as a vector of binaries)
    :return: last created operator
    """
    axes = []
    p = 0
    perm = []
    for i, r in enumerate(row):
        if r == -1:
            axes.append((p, i))
        else:
            p += 1
            perm.append((r, i))
    op = EinsumSubOp(len(row), 'expand_dims', op, axes=tuple(axes))
    yield op
    perm.sort()
    p = 0
    new_perm = numpy.arange(len(row))
    for i, r in enumerate(row):
        if r == -1:
            continue
        new_perm[perm[p][1]] = i
        p += 1
    if not is_transpose_identity(new_perm):
        op = EinsumSubOp(len(row), 'transpose', op, perm=tuple(new_perm))
        yield op


def _apply_squeeze_transpose(op, row_last, row_output):
    """
    Puts output dimension in the expected order.
    """
    perm = []
    sq = []
    for i, d in enumerate(row_output):
        if d == -1:
            sq.append(i)
        else:
            perm.append((d, i))
    perm.sort()
    new_perm = numpy.arange(len(row_last))
    p = 0
    for i, d in enumerate(row_output):
        if d == -1:
            continue
        new_perm[i] = perm[p][1]
        p += 1
    perm = [p[1] for p in perm]
    if not is_transpose_identity(new_perm):
        op = EinsumSubOp(len(row_last), 'transpose', op,
                         perm=tuple(new_perm))
        yield op
    if len(sq) > 0:
        op = EinsumSubOp(len(row_last), 'squeeze', op, axes=tuple(sq))
        yield op


def _apply_einsum_matmul(fd, op1, op2, axes, left, right, ndim,
                         op_matmul, row1, row2, verbose=False):
    """
    Decomposes the generic matrix multiplication into numpy operations
    depending on the operator to use for matrix multiplication
    *op_matmul* (see @see fn decompose_einsum_equation).
    """
    allowed = {'matmul', 'batch_dot', 'dot'}
    if op_matmul not in allowed:
        raise ValueError(  # pragma: no cover
            f"Unknown operator op_matmul={op_matmul!r} not in {allowed!r}.")
    if op_matmul == 'matmul':
        if verbose:  # pragma: no cover
            print(
                f"  -- MATMUL -> matmul axes={axes!r} left={left!r} right={right!r}")
        yield EinsumSubOp(fd, 'matmul', op1, op2,
                          axes=axes, left=left, right=right, ndim=ndim)

    elif len(axes) == 0 and len(set(left) & set(right)) == 0:
        if verbose:  # pragma: no cover
            print(
                f"  -- MATMUL -> mul axes={axes!r} left={left!r} right={right!r}")
        yield EinsumSubOp(fd, 'mul', op1, op2)

    elif (len(set(axes) & set(left)) == 0 and
            len(set(axes) & set(right)) == 0):

        # No intersection between axes and right: matrix multiplication
        if verbose:  # pragma: no cover
            print("  -- MATMUL -> batch_dot axes=%r left=%r right=%r"
                  "" % (axes, left, right))

        all_axes = set(left) | set(right) | set(axes)
        common_axes = list(set(left) & set(right))
        for i in range(ndim):
            if i not in all_axes:
                common_axes.append(i)
        common_axes.sort()

        # ReduceSum*
        has_dim = set(i for i in range(len(row1)) if row1[i] >= 0)
        right_no_left = (set(right) & has_dim) - \
            (set(right) & (set(left) | set(axes)))
        if right_no_left:
            if verbose:  # pragma: no cover
                print(
                    f'  -- MATMUL reduce1 has_dim={has_dim!r} axes={right_no_left!r}')
            op1 = EinsumSubOp(fd, 'reduce_sum_mm', op1, op2,
                              axes=tuple(sorted(right_no_left)))
            yield op1

        has_dim = set(i for i in range(len(row2)) if row2[i] >= 0)
        left_no_right = (set(left) & has_dim) - \
            (set(left) & (set(right) | set(axes)))
        if left_no_right:
            if verbose:  # pragma: no cover
                print(
                    f'  -- MATMUL reduce2 has_dim={has_dim!r} axes={left_no_right!r}')
            op2 = EinsumSubOp(fd, 'reduce_sum', op2,
                              axes=tuple(sorted(left_no_right)))
            yield op2

        # Transpose
        i_axes = [(-1 if i in common_axes
                   else (1 if i in axes else 0), i)
                  for i in range(ndim)]
        i_axes.sort()
        perm = [_[1] for _ in i_axes]
        perm_left = [i for i in range(len(perm)) if perm[i] in left]
        perm_right = [i for i in range(len(perm)) if perm[i] in right]
        if not is_transpose_identity(perm):
            op1 = EinsumSubOp(fd, 'transpose_mm', op1, op2, perm=tuple(perm))
            yield op1
            op2 = EinsumSubOp(fd, 'transpose', op2, perm=tuple(perm))
            yield op2

        # Reshape
        all_axes = list(range(0, ndim))
        new_axes = all_axes[-len(axes):] if len(axes) > 0 else []
        new_common_axes = all_axes[:len(common_axes)]
        not_in_both = []
        for i in range(0, ndim):
            if i not in left and i not in right and i not in common_axes:
                not_in_both.append(i)

        op = EinsumSubOp(fd, 'batch_dot', op1, op2,
                         batch_axes=tuple(new_common_axes),
                         keep_axes=None, sum_axes=tuple(new_axes),
                         left=tuple(perm_left), right=tuple(perm_right),
                         ndim=ndim)
        yield op

        # Transpose again
        ordered_axes = (common_axes +
                        list(i for i in left if i not in right) +
                        list(i for i in right if i not in left) +
                        not_in_both)
        rev_perm = [(a, i) for i, a in enumerate(ordered_axes)]
        rev_perm.sort()
        rev_perm = [p[1] for p in rev_perm]

        if not is_transpose_identity(rev_perm):
            op_unused = EinsumSubOp(fd, 'transpose_mm', op1,
                                    op, perm=tuple(rev_perm))
            yield op_unused
            op = EinsumSubOp(fd, 'transpose', op, perm=tuple(rev_perm))
            yield op
    else:
        raise NotImplementedError(  # pragma: no cover
            "axes and right or left have axes in common, "
            "axes=%r left=%r right=%r ndim=%r." % (
                axes, left, right, ndim))


def _decompose_einsum_equation_simple(equation, *shapes, verbose=False,
                                      op_matmul='matmul'):
    """
    Applies strategy `simple`, `numpy`
    defined in by function @see fn decompose_einsum_equation.

    :param op_matmul: which operator to use for matrix multiplication,
        a single operator *matmul*, or *batch_dot* with *transposes*,
        *reduce_sum*, or just *dot*
    """
    letters, mat, lengths, duplicates = analyse_einsum_equation(equation)
    if len(letters) != mat.shape[1]:
        raise RuntimeError(  # pragma: no cover
            f"Unexpected number of letters {letters!r}, shape={mat.shape!r}.")
    if len(shapes) == 0:
        shapes = [(2, ) * le for le in lengths[:-1]]
    _basic_verification(lengths, shapes, equation)

    # last_row, current_row (row = shape)
    rows = numpy.full((2, mat.shape[1]), -1)
    graph = GraphEinsumSubOp(letters, mat, lengths, duplicates)
    fd = mat.shape[1]
    if verbose:
        print(f"EQUATION={equation!r}")
        print(f"LETTERS={letters!r}", f"LENGTHS={lengths!r}")
        print(f"DUPLICATES={duplicates!r}")

    for i, sh in enumerate(shapes):
        if verbose:
            print()
            print("######### ROW %d shape=%r row=%r" % (i, sh, rows[1, :]))
        graph.append(i)

        # Input matrix aligned to the same dimensions.
        op = EinsumSubOp(fd, 'id', i)
        op.compute_output_row(rows[1, :], mat[i, :], verbose=verbose)
        marked = graph.append(op)

        duplicate = duplicates[i]
        if duplicate is not None:
            # Diagonal
            diag = []
            for _, v in duplicate.items():
                if len(v) == 1:
                    continue
                diag.append((v[0], tuple(v)))
            op = EinsumSubOp(fd, 'diagonal', op, diag=diag)
            op.compute_output_row(rows[1, :], mat[i, :], verbose=verbose)
            tr_row = rows[1, :]
            marked = graph.append(op)
        else:
            diag = None
            tr_row = mat[i]

        for op in _apply_transpose_reshape(op, tr_row):
            op.compute_output_row(rows[1, :], verbose=verbose)
            marked = graph.append(op)

        # Reduction? (a dimension not used later)
        red = []
        for d in range(0, mat.shape[1]):
            if (mat[i + 1:, d].max() == -1 and rows[1, d] != -1 and
                    rows[0, d] == -1):
                red.append(d)
        if len(red) > 0:
            if verbose:
                print("  -- REDUCE1 row=%d axes=%r" % (i, red))
                print(mat)
                print('  -')
                print(rows)
            op = EinsumSubOp(fd, 'reduce_sum',
                             graph.last_added_op, axes=tuple(red))
            op.compute_output_row(rows[1, :], verbose=verbose)
            marked = graph.append(op)

        if graph.last_op is not None:
            # Matrix multiplication?
            common_dims = []
            left = []
            right = []
            for d in range(0, mat.shape[1]):
                if rows[:, d].min() >= 0:
                    if mat[i + 1:, d].max() >= 0:
                        left.append(d)
                        right.append(d)
                    else:
                        common_dims.append(d)
                else:
                    if rows[0, d] >= 0:
                        left.append(d)
                    if rows[1, d] >= 0:
                        right.append(d)
            if verbose:
                print(f"  -- MATMUL common_dims={common_dims!r}")
                print(rows)
            for iop in _apply_einsum_matmul(
                    fd, graph.last_op, op, axes=tuple(common_dims),
                    left=tuple(left), right=tuple(right),
                    ndim=rows.shape[1], op_matmul=op_matmul,
                    row1=rows[0, :], row2=rows[1, :], verbose=verbose):
                op = iop
                op.compute_output_row(rows[0, :], rows[1, :],
                                      ab=True, verbose=verbose)
                marked = graph.append(op)

        # End
        graph.mark(i, marked)
        rows[0, :] = rows[1, :]

    # Final output
    if verbose:
        print()
        print(f"######### FIN row={rows[1, :]!r}")

    if mat[len(shapes), :].max() >= 0:
        rows[1, :] = mat[len(shapes), :]
        red = []
        for d in range(0, mat.shape[1]):
            if rows[0, d] > 0 and rows[1, d] == -1:
                red.append(d)
            elif rows[0, d] == -1 and rows[1, d] >= 0:
                raise RuntimeError(  # pragma: no cover
                    "Issue in equation %r, variable %d, last_result is %r, "
                    "output is %r." % (equation, d, rows[0, :], rows[1, :]))
        if len(red) > 0:
            if verbose:  # pragma: no cover
                print(f"-- REDUCE2 axes={red!r}")
                print(mat)
            op = EinsumSubOp(fd, 'reduce_sum', op, axes=tuple(red))
            graph.append(op)
            op.compute_output_row(rows[1, :], verbose=verbose)

        # Removes empty axes.
        for op in _apply_squeeze_transpose(op, rows[1, :], mat[len(shapes), :]):
            op.compute_output_row(rows[1, :], verbose=verbose)
            graph.append(op)
    return graph
